import struct

from ppc_asm.assembler import ppc
from ppc_asm.dol_file import DolEditor


def _split_int(value: int) -> tuple[int, int]:
    high = value >> 16
    low = struct.unpack("h", struct.pack("H", value & 0xFFFF))[0]

    if low < 0:
        # When low_value is considered negative, adding it will subtract the overflow bit
        # Add one to high_value to fix it
        high += 1

    return high, low


def load_32bit_int(
    dol: DolEditor, register: ppc.GeneralRegister, value: int, lis_address: int, addi_address: int
) -> None:
    """
    Patches the assembly instructions for loading a 32-bit int into a register, using two instructions.
    :return:
    """

    high_half, low_half = _split_int(value)

    # SetMLvlId argument
    dol.write_instructions(lis_address, [ppc.lis(register, high_half)])
    dol.write_instructions(addi_address, [ppc.addi(register, register, low_half)])



import hashlib
from pathlib import Path

import pytest

from open_prime_rando.echoes import legacy_patcher


@pytest.fixture(
    params=[False, True],
)
def is_legacy(request: pytest.FixtureRequest) -> bool:
    return request.param


def hash_all_paks(base_path: Path) -> dict[str, bytes]:
    return {
        pak_path.relative_to(base_path).as_posix(): hashlib.sha256(pak_path.read_bytes()).digest()
        for pak_path in base_path.rglob("*.pak")
    }


def test_ntsc_paks(prime2_iso_provider, tmp_path, test_files_dir) -> None:
    output_path = tmp_path.joinpath("out")
    configuration = test_files_dir.read_json("echoes", "door_lock.json")

    expected_hashes = {
        "files/AudioGrp.pak": b"U\xc6\xfe\\\xac\xcd\xdc\x02\x0c\xed\xdd\xb3\xfarbP"
        b"\x90\xac\x8d\x1b=\xd4\x9fl\x82\xc8\x12\xd2\x85X{+",
        "files/LogBook.pak": b"K=z\x02\x9c>\x8a\xfbN\xbe\xb6\xb5\xc8\xad\xba\xb8"
        b"\x99\xae\x85_\xab55\xfe3\x15\x0b\xdc\xd1\xc7\x02^",
        "files/Metroid1.pak": b"\xa7S\xf1'9^\x99\xc3y\x19\x88(k\x98.\xf8oh1\x0b\x1d:\x02\xc9\x0b\xa8%\xb7\xf6!a/",
        "files/Metroid2.pak": b'\xbdn\x83\xc6\x0c$)}\nv\xd1q"\x85\xa4\xc1\x99\xaa\xed\xc49-\xab\x08'
        b"\xf0\xef\xfa:\xc3]\xe8\xe3",
        "files/Metroid3.pak": b"\xab\xacT\xe2\xad\x03\xa3\xbecO#Ew!\x02\xfb\xec\x0eB\x13\xa4\x8b\xc96"
        b"\n\xb4\x16\x99\xeeC\xdbr",
        "files/Metroid4.pak": b"\xb3\xa1\xc3\xe0\x1d\xbdB\x7f\x96H\xa7_\xb8\x0e\xb2d\xe8\xe3\xde\xd6"
        b'#\x91L\xfaTI]"U\xd5\xea0',
        "files/Metroid5.pak": b"\xff\xe61\x08~u\xdf\xa4\xc8\xcf:\x8f\x89\xee\xa6{"
        b"b\xfe\xa7\xb8 r;{\x1e\xcf\xb4R\xc3\x8c\xac\x9a",
        "files/MiscData.pak": b"'\xbd\x82\x84\xdf\xf0\x88\x1d\xda\xe0|{\xafo\x80\xd5\xbc\xe4xR"
        b"\xb4W\xdb\x1eT\x00@\xba{m`]",
        "files/SamusGun.pak": b"\xdd\xa99\x0b\x14b\x97&\xee\xdf\nE\xd0GR`c\x85\xf7m;\xdbX^\xe4\x87;\\?\xc9\x08\x1f",
        "files/SamusGunLow.pak": b"o8\xb8\xab^\xc3y\xc7\xd8v\xf1?\xe7a\x9enIE\xb7\xa5"
        b"q\x9f\xa2\xf6\xc2\xc6m&\x80G\x8f\x83",
        "files/TestAnim.pak": b"\xc0n\xfa2\x87\x96o\xd9\xf8\nL\xc8\xd5vVJi\xe2\xa0\xdb"
        b"\xdbq\t\x7f\xaf\x96\x8d\x02\x1f\xe7\x07\x12",
    }

    legacy_patcher.patch_paks(
        file_provider=prime2_iso_provider,
        output_path=output_path,
        configuration=configuration,
    )
    hashes = hash_all_paks(output_path)
    assert hashes == expected_hashes


def test_pal_paks(pal_prime2_iso_provider, tmp_path, test_files_dir):
    output_path = tmp_path.joinpath("out")
    configuration = test_files_dir.read_json("echoes", "door_lock.json")

    legacy_patcher.patch_paks(
        file_provider=pal_prime2_iso_provider,
        output_path=output_path,
        configuration=configuration,
    )
    assert len(list(output_path.rglob("*.pak"))) == 11



from retro_data_structures.game_check import Game

from open_prime_rando.dol_patching.all_prime_dol_patches import PowerupFunctionsAddresses
from open_prime_rando.dol_patching.corruption.dol_patches import CorruptionDolVersion

ALL_VERSIONS = [
    CorruptionDolVersion(
        game=Game.CORRUPTION,
        description="Wii NTSC",
        build_string_address=0x805822B0,
        build_string=b"!#$MetroidBuildInfo!#$2007/07/27 13:13 Build v3.436 (MP3)",
        sda2_base=0x806869C0,
        sda13_base=0x806801C0,
        game_state_pointer=0x8067DC0C,
        cplayer_vtable=0x80592C78,
        cstate_manager_global=0x805C4F70,
        string_display=None,
        # string_display=StringDisplayPatchAddresses(
        #     update_hint_state=None,
        #     message_receiver_string_ref=0x805a4200,
        #     wstring_constructor=None,
        #     display_hud_memo=0x801c6480,
        #     max_message_size=200,
        # ),
        powerup_functions=PowerupFunctionsAddresses(
            add_power_up=0x8019111C,
            incr_pickup=0x801913A0,
            decr_pickup=0x8019151C,
        ),
    ),
    CorruptionDolVersion(
        game=Game.CORRUPTION,
        description="Wii PAL",
        build_string_address=0x805843A8,
        build_string=b"!#$MetroidBuildInfo!#$2007/08/24 16:52 Build v3.453 (mp3)",
        sda2_base=0x80688FE0,
        sda13_base=0x806827C0,
        game_state_pointer=0x80680234,
        cplayer_vtable=0x80595238,
        cstate_manager_global=0x805C7570,
        string_display=None,
        powerup_functions=PowerupFunctionsAddresses(
            add_power_up=0x80191E2C,
            incr_pickup=0x801920B0,
            decr_pickup=0x8019222C,
        ),
    ),
    CorruptionDolVersion(
        game=Game.CORRUPTION,
        description="Wii NTSC-J",
        build_string_address=0x80587C2C,
        build_string=b"!#$MetroidBuildInfo!#$2007/11/12 14:15 Build v3.495 (jpn)",
        sda2_base=0x8068C840,
        sda13_base=0x80686000,
        game_state_pointer=0x80683A7C,
        cplayer_vtable=0x80598650,
        cstate_manager_global=0x805CAA30,
        string_display=None,
        powerup_functions=PowerupFunctionsAddresses(
            add_power_up=0x801932D0,
            incr_pickup=0x80193554,
            decr_pickup=0x801936D0,
        ),
    ),
]



from unittest.mock import MagicMock

import pytest_mock
from retro_data_structures.game_check import Game

from open_prime_rando.dol_patching.echoes import dol_patcher, dol_versions


def test_apply_patches(mocker: pytest_mock.MockerFixture):
    # Setup
    patches_data = MagicMock()
    version_patches = dol_versions.ALL_VERSIONS[0]
    mock_find_version_for_dol = mocker.patch(
        "open_prime_rando.dol_patching.dol_version.find_version_for_dol",
        return_value=version_patches,
        autospec=True,
    )
    dol_file = MagicMock()

    mock_apply_remote_exec = mocker.patch(
        "open_prime_rando.dol_patching.all_prime_dol_patches.apply_remote_execution_patch", autospec=True
    )
    mock_apply_etank_capacity: MagicMock = mocker.patch(
        "open_prime_rando.dol_patching.all_prime_dol_patches.apply_energy_tank_capacity_patch", autospec=True
    )

    mock_apply_game_options: MagicMock = mocker.patch(
        "open_prime_rando.dol_patching.echoes.dol_patches.apply_game_options_patch", autospec=True
    )
    mock_apply_beam_cost_patch: MagicMock = mocker.patch(
        "open_prime_rando.dol_patching.echoes.dol_patches.apply_beam_cost_patch", autospec=True
    )
    mock_apply_starting_visor_patch: MagicMock = mocker.patch(
        "open_prime_rando.dol_patching.echoes.dol_patches.apply_starting_visor_patch", autospec=True
    )
    mock_apply_fixes: MagicMock = mocker.patch(
        "open_prime_rando.dol_patching.echoes.dol_patches.apply_fixes", autospec=True
    )
    mock_apply_unvisited_room_names: MagicMock = mocker.patch(
        "open_prime_rando.dol_patching.echoes.dol_patches.apply_unvisited_room_names", autospec=True
    )
    mock_apply_teleporter_sounds: MagicMock = mocker.patch(
        "open_prime_rando.dol_patching.echoes.dol_patches.apply_teleporter_sounds", autospec=True
    )

    # Run
    dol_patcher.apply_patches(dol_file, patches_data)

    # Assert
    mock_find_version_for_dol.assert_called_once_with(dol_file, dol_versions.ALL_VERSIONS)
    mock_apply_remote_exec.assert_called_once_with(Game.ECHOES, version_patches.string_display, dol_file)
    mock_apply_game_options.assert_called_once_with(
        version_patches.game_options_constructor_address, patches_data.user_preferences, dol_file
    )
    mock_apply_etank_capacity.assert_called_once_with(
        version_patches.health_capacity, patches_data.energy_per_tank, dol_file
    )
    mock_apply_beam_cost_patch.assert_called_once_with(
        version_patches.beam_cost_addresses, patches_data.beam_configurations, dol_file
    )
    mock_apply_starting_visor_patch.assert_called_once_with(
        version_patches.starting_beam_visor,
        patches_data.default_items,
        dol_file,
    )
    mock_apply_fixes.assert_called_once_with(version_patches, dol_file)
    mock_apply_unvisited_room_names.assert_called_once_with(
        version_patches, dol_file, patches_data.unvisited_room_names
    )
    mock_apply_teleporter_sounds.assert_called_once_with(version_patches, dol_file, patches_data.teleporter_sounds)



import dataclasses
import struct
import uuid

from ppc_asm import assembler
from ppc_asm.assembler import custom_ppc
from ppc_asm.assembler.ppc import *  # noqa: F403
from ppc_asm.dol_file import DolEditor
from retro_data_structures.game_check import Game

from open_prime_rando.dol_patching.dol_version import DolVersion

ARTIFACT_ITEMS = list(range(29, 41))

# ruff: noqa: F405


@dataclasses.dataclass(frozen=True)
class StringDisplayPatchAddresses:
    update_hint_state: int
    message_receiver_string_ref: int
    wstring_constructor: int
    display_hud_memo: int
    max_message_size: int


@dataclasses.dataclass(frozen=True)
class HealthCapacityAddresses:
    base_health_capacity: int
    energy_tank_capacity: int


@dataclasses.dataclass(frozen=True)
class DangerousEnergyTankAddresses:
    small_number_float: int
    incr_pickup: int


@dataclasses.dataclass(frozen=True)
class PowerupFunctionsAddresses:
    """
    add_power_up: Change the given item capacity
    incr_pickup: Increments the given item amount
    decr_pickup: Decrements the given item amount
    """

    add_power_up: int
    incr_pickup: int
    decr_pickup: int


@dataclasses.dataclass(frozen=True)
class BasePrimeDolVersion(DolVersion):
    game_state_pointer: int
    cplayer_vtable: int
    cstate_manager_global: int
    string_display: StringDisplayPatchAddresses
    powerup_functions: PowerupFunctionsAddresses


_registers_to_save = 2
_remote_execution_stack_size = 0x30 + (_registers_to_save * 4)
_remote_execution_max_byte_count = 420  # Prime 1 0-00 is 444, Echoes NTSC is 464


def remote_execution_patch_start(game: Game) -> list[BaseInstruction]:
    return_code = remote_execution_cleanup_and_return()
    return_code_byte_count = assembler.byte_count(return_code)

    if game == Game.PRIME:
        cutscene_check = [
            lwz(r3, 0x870, r31),
            cmpwi(r3, 0),
            beq(-0x8 - return_code_byte_count, True),
            lwz(r0, 0x8, r3),
        ]
    elif game == Game.ECHOES:
        cutscene_check = [
            lwz(r0, 0x16FC, r31),
        ]
    else:
        raise ValueError(f"Unsupported game: {game}")

    intro = [
        # setup stack
        stwu(r1, -(_remote_execution_stack_size - 4), r1),
        mfspr(r0, LR),
        stw(r0, _remote_execution_stack_size, r1),
        stmw(GeneralRegister(32 - _registers_to_save), _remote_execution_stack_size - 4 - _registers_to_save * 4, r1),
        or_(r31, r3, r3),
        # return if no pending op
        lbz(r4, 0x2, r31),
        cmpwi(r4, 0x0),
        bne((len(return_code) + 1) * 4, relative=True),
        # clean return if flag is not set
        *return_code,
        # if camera count > 0 then we're in a cutscene so we return as we don't want
        # to handle receiving items during a cutscene
        *cutscene_check,
        cmpwi(r0, 0),
        bgt(-assembler.byte_count(cutscene_check) - 4 - return_code_byte_count, True),
    ]

    num_bytes_to_invalidate = _remote_execution_max_byte_count - assembler.byte_count(intro)
    # Our loop end condition depends on this value being a multiple of 32, greater than 0
    num_bytes_to_invalidate = ((num_bytes_to_invalidate // 32) + 1) * 32

    return [
        *intro,
        # fetch the instructions again, since they're being overwritten externally
        # this clears Dolphin's JIT cache
        custom_ppc.load_current_address(r30, 8),
        li(r4, num_bytes_to_invalidate),
        icbi(4, 30),  # invalidate using r30 + r4
        cmpwi(r4, 0x0),
        addi(r4, r4, -32),
        bne(-3 * 4, relative=True),
        sync(),
        isync(),
    ]


def remote_execution_clear_pending_op() -> list[BaseInstruction]:
    return [
        # set no pending op
        li(r6, 0x0),
        stb(r6, 0x2, r31),
    ]


def remote_execution_cleanup_and_return() -> list[BaseInstruction]:
    return [
        # restore value of top registers
        lmw(GeneralRegister(32 - _registers_to_save), _remote_execution_stack_size - 4 - _registers_to_save * 4, r1),
        lwz(r0, _remote_execution_stack_size, r1),
        mtspr(LR, r0),
        addi(r1, r1, _remote_execution_stack_size - 4),
        blr(),
    ]


def call_display_hud_patch(patch_addresses: StringDisplayPatchAddresses) -> list[BaseInstruction]:
    # r31 = CStateManager
    return [
        # setup CHUDMemoParms
        lis(r5, 0x4100),  # 8.0f
        li(r6, 0x0),
        li(r7, 0x1),
        li(r9, 0x9),
        stw(r5, 0x10, r1),  # display time (seconds)
        stb(r7, 0x14, r1),  # clear memo window
        stb(r6, 0x15, r1),  # fade out only
        stb(r6, 0x16, r1),  # hint memo
        stb(r7, 0x17, r1),  # fade in text
        stw(r9, 0x18, r1),  # unk
        # setup wstring
        addi(r3, r1, 0x1C),
        custom_ppc.load_unsigned_32bit(r4, patch_addresses.message_receiver_string_ref),
        bl(patch_addresses.wstring_constructor),
        # r4 = CHUDMemoParms
        addi(r4, r1, 0x10),
        bl(patch_addresses.display_hud_memo),
    ]


def _load_player_state(game: Game, target_register: GeneralRegister, state_mgr: GeneralRegister = r31):
    if game == Game.PRIME:
        return [
            lwz(target_register, 0x8B8, state_mgr),
            lwz(target_register, 0, target_register),
        ]
    elif game == Game.ECHOES:
        return [
            lwz(target_register, 0x150C, state_mgr),
        ]
    else:
        raise ValueError(f"Unsupported game {game}")


def adjust_item_amount_and_capacity_patch(
    patch_addresses: PowerupFunctionsAddresses,
    game: Game,
    item_id: int,
    delta: int,
) -> list[BaseInstruction]:
    # r31 = CStateManager
    if game == Game.PRIME and item_id in ARTIFACT_ITEMS:
        return increment_item_capacity_patch(patch_addresses, game, item_id, delta)

    if delta >= 0:
        return [
            *increment_item_capacity_patch(patch_addresses, game, item_id, delta),
            *adjust_item_amount_patch(patch_addresses, game, item_id, delta),
        ]
    else:
        return [
            *adjust_item_amount_patch(patch_addresses, game, item_id, delta),
            *increment_item_capacity_patch(patch_addresses, game, item_id, delta),
        ]


def increment_item_capacity_patch(
    patch_addresses: PowerupFunctionsAddresses,
    game: Game,
    item_id: int,
    delta: int = 1,
) -> list[BaseInstruction]:
    return [
        *_load_player_state(game, r3, r31),
        li(r4, item_id),
        li(r5, delta),
        bl(patch_addresses.add_power_up),
    ]


def adjust_item_amount_patch(
    patch_addresses: PowerupFunctionsAddresses,
    game: Game,
    item_id: int,
    delta: int,
) -> list[BaseInstruction]:
    return [
        *_load_player_state(game, r3, r31),
        li(r4, item_id),
        li(r5, abs(delta)),
        bl(patch_addresses.incr_pickup if delta >= 0 else patch_addresses.decr_pickup),
    ]


def remote_execution_patch(game: Game):
    return [
        *remote_execution_patch_start(game),
        *remote_execution_clear_pending_op(),
        *remote_execution_cleanup_and_return(),
    ]


def apply_remote_execution_patch(game: Game, patch_addresses: StringDisplayPatchAddresses, dol_editor: DolEditor):
    dol_editor.write_instructions(patch_addresses.update_hint_state, remote_execution_patch(game))


def create_remote_execution_body(
    game: Game, patch_addresses: StringDisplayPatchAddresses, instructions: list[BaseInstruction]
) -> tuple[int, bytes]:
    """
    Return the address and the bytes for executing the given instructions via remote code execution.
    """
    update_hint_state = patch_addresses.update_hint_state

    remote_start_instructions = remote_execution_patch_start(game)
    remote_start_byte_count = assembler.byte_count(remote_start_instructions)

    body_address = update_hint_state + remote_start_byte_count
    body_instructions = list(instructions)
    body_instructions.extend(remote_execution_clear_pending_op())
    body_instructions.extend(remote_execution_cleanup_and_return())
    body_bytes = bytes(assembler.assemble_instructions(body_address, body_instructions))

    if len(body_bytes) > _remote_execution_max_byte_count - remote_start_byte_count:
        raise ValueError(
            f"Received {len(body_instructions)} instructions with total {len(body_bytes)} bytes, "
            f"but limit is {_remote_execution_max_byte_count - remote_start_byte_count}."
        )

    return body_address, body_bytes


def apply_energy_tank_capacity_patch(
    patch_addresses: HealthCapacityAddresses, energy_per_tank: int, dol_editor: DolEditor
):
    """
    Patches the base health capacity and the energy tank capacity with matching values.
    """
    tank_capacity = float(energy_per_tank)

    dol_editor.write(patch_addresses.base_health_capacity, struct.pack(">f", tank_capacity - 1))
    dol_editor.write(patch_addresses.energy_tank_capacity, struct.pack(">f", tank_capacity))


def apply_reverse_energy_tank_heal_patch(
    sd2_base: int,
    addresses: DangerousEnergyTankAddresses,
    active: bool,
    game: Game,
    dol_editor: DolEditor,
):
    if game == Game.ECHOES:
        health_offset = 0x14
        refill_item = 0x29
        patch_offset = 0x90

    elif game == Game.CORRUPTION:
        health_offset = 0xC
        refill_item = 0x12
        patch_offset = 0x138

    else:
        raise ValueError(f"Unsupported game: {game}")

    if active:
        patch = [
            lfs(f0, (addresses.small_number_float - sd2_base), r2),
            stfs(f0, health_offset, r30),
            ori(r0, r0, 0),
            ori(r0, r0, 0),
        ]
    else:
        patch = [
            or_(r3, r30, r30),
            li(r4, refill_item),
            li(r5, 9999),
            bl(addresses.incr_pickup),
        ]

    dol_editor.write_instructions(addresses.incr_pickup + patch_offset, patch)


def apply_build_info_patch(dol_file: DolEditor, uid: uuid.UUID, version: DolVersion):
    dol_file.write(
        version.build_string_address + 6,
        uid.bytes,
    )



import dataclasses

from open_prime_rando.dol_patching.all_prime_dol_patches import BasePrimeDolVersion


@dataclasses.dataclass(frozen=True)
class CorruptionDolVersion(BasePrimeDolVersion):
    pass



import argparse
from pathlib import Path

from open_prime_rando.echoes import patcher
from open_prime_rando.echoes.rando_configuration import RandoConfiguration


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--input-iso", required=True, type=Path, help="Path to a ISO to randomize")
    parser.add_argument("--output-iso", required=True, type=Path, help="Path to write the modified ISO.")

    subparser = parser.add_subparsers(dest="command")
    rando = subparser.add_parser("randomizer")
    rando.add_argument("--input-json", type=Path, required=True, help="Path to the configuration json.")
    return parser


def main(argv: list[str]) -> None:
    parser = create_parser()
    args = parser.parse_args(argv)

    print(args)

    if args.command == "randomizer":
        configuration_path: Path = args.input_json
        configuration = RandoConfiguration.model_validate_json(
            configuration_path.read_text(),
            strict=True,
        )

        patcher.patch_iso(
            args.input_iso,
            args.output_iso,
            configuration,
        )
    else:
        parser.print_help()
        raise SystemExit(1)



import struct
from enum import IntEnum
from typing import NamedTuple


class DoorIconColors(NamedTuple):
    surface_color: int
    outline_color: int | None = None  # unused


class DoorMapIcon(IntEnum):
    # vanilla
    Normal = 0
    Missile = 1
    Dark = 2
    Annihilator = 3
    Light = 4
    SuperMissile = 5
    SeekerMissile = 6
    PowerBomb = 7

    # planetary energy
    AgonEnergy = -1
    TorvusEnergy = -2
    SanctuaryEnergy = -3

    # visors
    ScanVisor = -4
    DarkVisor = -5
    EchoVisor = -6

    # misc
    Disabled = -7
    ScrewAttack = -8
    Bomb = -9
    Boost = -10
    Grapple = -11

    @property
    def colors(self) -> DoorIconColors:
        return _ALL_COLORS.get(self, DoorIconColors(0xFF00FFFF))

    @staticmethod
    def get_surface_colors_as_bytes() -> bytes:
        colors = [icon.colors.surface_color for icon in sorted(DoorMapIcon)]
        return struct.pack(">" + "L" * len(colors), *colors)

    @staticmethod
    def get_door_index_bounds() -> tuple[int, int]:
        return min(*DoorMapIcon), max(*DoorMapIcon)


_ALL_COLORS = {
    DoorMapIcon.Normal: DoorIconColors(0x3379BFFF),
    DoorMapIcon.Missile: DoorIconColors(0xD63333FF),
    DoorMapIcon.Dark: DoorIconColors(0x000000FF, 0x898989FF),
    DoorMapIcon.Annihilator: DoorIconColors(0x4B4B4BFF),
    DoorMapIcon.Light: DoorIconColors(0xFFFFFFFF),
    DoorMapIcon.SuperMissile: DoorIconColors(0x50A148FF),
    DoorMapIcon.SeekerMissile: DoorIconColors(0x794F77FF),
    DoorMapIcon.PowerBomb: DoorIconColors(0xEAE50BFF),
    DoorMapIcon.AgonEnergy: DoorIconColors(0xA45600FF),
    DoorMapIcon.TorvusEnergy: DoorIconColors(0x4E9761FF),
    DoorMapIcon.SanctuaryEnergy: DoorIconColors(0x56789DFF),
    DoorMapIcon.ScanVisor: DoorIconColors(0x007F7FFF),
    DoorMapIcon.DarkVisor: DoorIconColors(0x660000FF),
    DoorMapIcon.EchoVisor: DoorIconColors(0xCC7A00FF),
    DoorMapIcon.ScrewAttack: DoorIconColors(0xED94D4FF),
    DoorMapIcon.Bomb: DoorIconColors(0x55557FFF),
    DoorMapIcon.Boost: DoorIconColors(0xFF5500FF),
    # DoorMapIcon.Grapple: None,
    DoorMapIcon.Disabled: DoorIconColors(0x202020FF),
}



# Generated by asset_id_files.py

ABANDONED_WORKSITE_MREA = 0x54DAEF57
AMMO_STATION_MREA = 0x8697F62A
BROODING_GROUND_MREA = 0xDCEF49B9
CACHE_A_MREA = 0x0CAAEEC6
CACHE_B_MREA = 0x007551F9
CATACOMBS_MREA = 0xFB628DCB
CATACOMBS_ACCESS_MREA = 0xB2CAA0C9
CONTROLLER_ACCESS_MREA = 0x2AF65B49
CRYPT_MREA = 0xC8A84B98
CRYPT_TUNNEL_MREA = 0x4DC421AD
DARK_ARENA_TUNNEL_MREA = 0x390DFCB9
DARK_CONTROLLER_ACCESS_MREA = 0x11C4526B
DARK_FALLS_MREA = 0x2E259879
DARK_FORGOTTEN_BRIDGE_MREA = 0x0253782D
DARK_TORVUS_ARENA_MREA = 0xA49A572F
DARK_TORVUS_ENERGY_CONTROLLER_MREA = 0x5ED7CD19
DARK_TORVUS_TEMPLE_MREA = 0x3663B195
DARK_TORVUS_TEMPLE_ACCESS_MREA = 0xE597E4F9
DUNGEON_MREA = 0xBBE4B3AE
FORGOTTEN_BRIDGE_MREA = 0x09DECF21
FORTRESS_TRANSPORT_ACCESS_MREA = 0x63190A61
GATHERING_ACCESS_MREA = 0x530DFF81
GATHERING_HALL_MREA = 0x3501473A
GLOOM_VISTA_MREA = 0x954888C7
GREAT_BRIDGE_MREA = 0xE3D5A55E
GROVE_ACCESS_MREA = 0x46644C45
HYDROCHAMBER_STORAGE_MREA = 0x96A6E17E
HYDRODYNAMO_SHAFT_MREA = 0x308351E5
HYDRODYNAMO_STATION_MREA = 0xD5B5B066
MAIN_ENERGY_CONTROLLER_MREA = 0xA6803305
MAIN_HYDROCHAMBER_MREA = 0xCEBAC0BD
MEDITATION_VISTA_MREA = 0x5DFA32DB
PATH_OF_ROOTS_MREA = 0x7448931C
PLAZA_ACCESS_MREA = 0xC1C28706
POISONED_BOG_MREA = 0x14E2C3A6
POLLUTED_MIRE_MREA = 0x839DC4F6
PORTAL_CHAMBER_DARK_MREA = 0xAE1E1339
PORTAL_CHAMBER_LIGHT_MREA = 0xF3EE585F
PUTRID_ALCOVE_MREA = 0xCBC72D79
RUINED_ALCOVE_MREA = 0xD37C2414
SACRIFICIAL_CHAMBER_MREA = 0x629790F4
SACRIFICIAL_CHAMBER_TUNNEL_MREA = 0x65E571CD
SAVE_STATION_1_MREA = 0x9A777442
SAVE_STATION_2_MREA = 0x13254EC6
SAVE_STATION_A_MREA = 0xDB6E4075
SAVE_STATION_B_MREA = 0xCB165BD8
TEMPLE_ACCESS_MREA = 0x0740B4D3
TEMPLE_TRANSPORT_ACCESS_MREA = 0x27D2C898
TORVUS_ENERGY_CONTROLLER_MREA = 0x133BF5B8
TORVUS_GROVE_MREA = 0xE838FA18
TORVUS_LAGOON_MREA = 0x7E0BE699
TORVUS_MAP_STATION_MREA = 0x282B587F
TORVUS_PLAZA_MREA = 0x75F158FF
TORVUS_TEMPLE_MREA = 0xE640CE96
TRAINING_ACCESS_MREA = 0x77AF9FD1
TRAINING_CHAMBER_MREA = 0x4BB5AE60
TRANSIT_TUNNEL_EAST_MREA = 0x66F6300E
TRANSIT_TUNNEL_SOUTH_MREA = 0xB7259AA6
TRANSIT_TUNNEL_WEST_MREA = 0xE150FB4D
TRANSPORT_TO_AGON_WASTES_MREA = 0xCF659F4E
TRANSPORT_TO_SANCTUARY_FORTRESS_MREA = 0xBF0EE428
TRANSPORT_TO_TEMPLE_GROUNDS_MREA = 0x6F6515F2
UNDERGROUND_TRANSPORT_MREA = 0x9258DC82
UNDERGROUND_TUNNEL_MREA = 0x80E67F90
UNDERTEMPLE_MREA = 0x85A3801F
UNDERTEMPLE_ACCESS_MREA = 0x29B990F8
UNDERTEMPLE_SHAFT_MREA = 0xB2D3E26C
UNDERTRANSIT_ONE_MREA = 0x5CD4C2F9
UNDERTRANSIT_TWO_MREA = 0x4BFCA639
VENOMOUS_POND_MREA = 0x858EECD3

NAME_TO_ID_MREA = {
    "Abandoned Worksite": 0x54DAEF57,
    "Ammo Station": 0x8697F62A,
    "Brooding Ground": 0xDCEF49B9,
    "Cache A": 0x0CAAEEC6,
    "Cache B": 0x007551F9,
    "Catacombs": 0xFB628DCB,
    "Catacombs Access": 0xB2CAA0C9,
    "Controller Access": 0x2AF65B49,
    "Crypt": 0xC8A84B98,
    "Crypt Tunnel": 0x4DC421AD,
    "Dark Arena Tunnel": 0x390DFCB9,
    "Dark Controller Access": 0x11C4526B,
    "Dark Falls": 0x2E259879,
    "Dark Forgotten Bridge": 0x0253782D,
    "Dark Torvus Arena": 0xA49A572F,
    "Dark Torvus Energy Controller": 0x5ED7CD19,
    "Dark Torvus Temple": 0x3663B195,
    "Dark Torvus Temple Access": 0xE597E4F9,
    "Dungeon": 0xBBE4B3AE,
    "Forgotten Bridge": 0x09DECF21,
    "Fortress Transport Access": 0x63190A61,
    "Gathering Access": 0x530DFF81,
    "Gathering Hall": 0x3501473A,
    "Gloom Vista": 0x954888C7,
    "Great Bridge": 0xE3D5A55E,
    "Grove Access": 0x46644C45,
    "Hydrochamber Storage": 0x96A6E17E,
    "Hydrodynamo Shaft": 0x308351E5,
    "Hydrodynamo Station": 0xD5B5B066,
    "Main Energy Controller": 0xA6803305,
    "Main Hydrochamber": 0xCEBAC0BD,
    "Meditation Vista": 0x5DFA32DB,
    "Path of Roots": 0x7448931C,
    "Plaza Access": 0xC1C28706,
    "Poisoned Bog": 0x14E2C3A6,
    "Polluted Mire": 0x839DC4F6,
    "Portal Chamber (Dark)": 0xAE1E1339,
    "Portal Chamber (Light)": 0xF3EE585F,
    "Putrid Alcove": 0xCBC72D79,
    "Ruined Alcove": 0xD37C2414,
    "Sacrificial Chamber": 0x629790F4,
    "Sacrificial Chamber Tunnel": 0x65E571CD,
    "Save Station 1": 0x9A777442,
    "Save Station 2": 0x13254EC6,
    "Save Station A": 0xDB6E4075,
    "Save Station B": 0xCB165BD8,
    "Temple Access": 0x0740B4D3,
    "Temple Transport Access": 0x27D2C898,
    "Torvus Energy Controller": 0x133BF5B8,
    "Torvus Grove": 0xE838FA18,
    "Torvus Lagoon": 0x7E0BE699,
    "Torvus Map Station": 0x282B587F,
    "Torvus Plaza": 0x75F158FF,
    "Torvus Temple": 0xE640CE96,
    "Training Access": 0x77AF9FD1,
    "Training Chamber": 0x4BB5AE60,
    "Transit Tunnel East": 0x66F6300E,
    "Transit Tunnel South": 0xB7259AA6,
    "Transit Tunnel West": 0xE150FB4D,
    "Transport to Agon Wastes": 0xCF659F4E,
    "Transport to Sanctuary Fortress": 0xBF0EE428,
    "Transport to Temple Grounds": 0x6F6515F2,
    "Underground Transport": 0x9258DC82,
    "Underground Tunnel": 0x80E67F90,
    "Undertemple": 0x85A3801F,
    "Undertemple Access": 0x29B990F8,
    "Undertemple Shaft": 0xB2D3E26C,
    "Undertransit One": 0x5CD4C2F9,
    "Undertransit Two": 0x4BFCA639,
    "Venomous Pond": 0x858EECD3,
}
# Generated by asset_id_files.py

ABANDONED_WORKSITE_MAPA = 0x7D0F146A
AMMO_STATION_MAPA = 0xAF420D17
BROODING_GROUND_MAPA = 0xF53AB284
CACHE_A_MAPA = 0x257F15FB
CACHE_B_MAPA = 0x29A0AAC4
CATACOMBS_MAPA = 0xD2B776F6
CATACOMBS_ACCESS_MAPA = 0x9B1F5BF4
CONTROLLER_ACCESS_MAPA = 0x0323A074
CRYPT_MAPA = 0xE17DB0A5
CRYPT_TUNNEL_MAPA = 0x6411DA90
DARK_ARENA_TUNNEL_MAPA = 0x10D80784
DARK_CONTROLLER_ACCESS_MAPA = 0x3811A956
DARK_FALLS_MAPA = 0x07F06344
DARK_FORGOTTEN_BRIDGE_MAPA = 0x2B868310
DARK_TORVUS_ARENA_MAPA = 0x8D4FAC12
DARK_TORVUS_ENERGY_CONTROLLER_MAPA = 0x77023624
DARK_TORVUS_TEMPLE_MAPA = 0x1FB64AA8
DARK_TORVUS_TEMPLE_ACCESS_MAPA = 0xCC421FC4
DUNGEON_MAPA = 0x92314893
FORGOTTEN_BRIDGE_MAPA = 0x200B341C
FORTRESS_TRANSPORT_ACCESS_MAPA = 0x4ACCF15C
GATHERING_ACCESS_MAPA = 0x7AD804BC
GATHERING_HALL_MAPA = 0x1CD4BC07
GLOOM_VISTA_MAPA = 0xBC9D73FA
GREAT_BRIDGE_MAPA = 0xCA005E63
GROVE_ACCESS_MAPA = 0x6FB1B778
HYDROCHAMBER_STORAGE_MAPA = 0xBF731A43
HYDRODYNAMO_SHAFT_MAPA = 0x1956AAD8
HYDRODYNAMO_STATION_MAPA = 0xFC604B5B
MAIN_ENERGY_CONTROLLER_MAPA = 0x8F55C838
MAIN_HYDROCHAMBER_MAPA = 0xE76F3B80
MEDITATION_VISTA_MAPA = 0x742FC9E6
PATH_OF_ROOTS_MAPA = 0x5D9D6821
PLAZA_ACCESS_MAPA = 0xE8177C3B
POISONED_BOG_MAPA = 0x3D37389B
POLLUTED_MIRE_MAPA = 0xAA483FCB
PORTAL_CHAMBER_DARK_MAPA = 0x87CBE804
PORTAL_CHAMBER_LIGHT_MAPA = 0xDA3BA362
PUTRID_ALCOVE_MAPA = 0xE212D644
RUINED_ALCOVE_MAPA = 0xFAA9DF29
SACRIFICIAL_CHAMBER_MAPA = 0x4B426BC9
SACRIFICIAL_CHAMBER_TUNNEL_MAPA = 0x4C308AF0
SAVE_STATION_1_MAPA = 0xB3A28F7F
SAVE_STATION_2_MAPA = 0x3AF0B5FB
SAVE_STATION_A_MAPA = 0xF2BBBB48
SAVE_STATION_B_MAPA = 0xE2C3A0E5
TEMPLE_ACCESS_MAPA = 0x2E954FEE
TEMPLE_TRANSPORT_ACCESS_MAPA = 0x0E0733A5
TORVUS_ENERGY_CONTROLLER_MAPA = 0x3AEE0E85
TORVUS_GROVE_MAPA = 0xC1ED0125
TORVUS_LAGOON_MAPA = 0x57DE1DA4
TORVUS_MAP_STATION_MAPA = 0x01FEA342
TORVUS_PLAZA_MAPA = 0x5C24A3C2
TORVUS_TEMPLE_MAPA = 0xCF9535AB
TRAINING_ACCESS_MAPA = 0x5E7A64EC
TRAINING_CHAMBER_MAPA = 0x6260555D
TRANSIT_TUNNEL_EAST_MAPA = 0x4F23CB33
TRANSIT_TUNNEL_SOUTH_MAPA = 0x9EF0619B
TRANSIT_TUNNEL_WEST_MAPA = 0xC8850070
TRANSPORT_TO_AGON_WASTES_MAPA = 0xE6B06473
TRANSPORT_TO_SANCTUARY_FORTRESS_MAPA = 0x96DB1F15
TRANSPORT_TO_TEMPLE_GROUNDS_MAPA = 0x46B0EECF
UNDERGROUND_TRANSPORT_MAPA = 0xBB8D27BF
UNDERGROUND_TUNNEL_MAPA = 0xA93384AD
UNDERTEMPLE_MAPA = 0xAC767B22
UNDERTEMPLE_ACCESS_MAPA = 0x006C6BC5
UNDERTEMPLE_SHAFT_MAPA = 0x9B061951
UNDERTRANSIT_ONE_MAPA = 0x750139C4
UNDERTRANSIT_TWO_MAPA = 0x62295D04
VENOMOUS_POND_MAPA = 0xAC5B17EE

NAME_TO_ID_MAPA = {
    "Abandoned Worksite": 0x7D0F146A,
    "Ammo Station": 0xAF420D17,
    "Brooding Ground": 0xF53AB284,
    "Cache A": 0x257F15FB,
    "Cache B": 0x29A0AAC4,
    "Catacombs": 0xD2B776F6,
    "Catacombs Access": 0x9B1F5BF4,
    "Controller Access": 0x0323A074,
    "Crypt": 0xE17DB0A5,
    "Crypt Tunnel": 0x6411DA90,
    "Dark Arena Tunnel": 0x10D80784,
    "Dark Controller Access": 0x3811A956,
    "Dark Falls": 0x07F06344,
    "Dark Forgotten Bridge": 0x2B868310,
    "Dark Torvus Arena": 0x8D4FAC12,
    "Dark Torvus Energy Controller": 0x77023624,
    "Dark Torvus Temple": 0x1FB64AA8,
    "Dark Torvus Temple Access": 0xCC421FC4,
    "Dungeon": 0x92314893,
    "Forgotten Bridge": 0x200B341C,
    "Fortress Transport Access": 0x4ACCF15C,
    "Gathering Access": 0x7AD804BC,
    "Gathering Hall": 0x1CD4BC07,
    "Gloom Vista": 0xBC9D73FA,
    "Great Bridge": 0xCA005E63,
    "Grove Access": 0x6FB1B778,
    "Hydrochamber Storage": 0xBF731A43,
    "Hydrodynamo Shaft": 0x1956AAD8,
    "Hydrodynamo Station": 0xFC604B5B,
    "Main Energy Controller": 0x8F55C838,
    "Main Hydrochamber": 0xE76F3B80,
    "Meditation Vista": 0x742FC9E6,
    "Path of Roots": 0x5D9D6821,
    "Plaza Access": 0xE8177C3B,
    "Poisoned Bog": 0x3D37389B,
    "Polluted Mire": 0xAA483FCB,
    "Portal Chamber (Dark)": 0x87CBE804,
    "Portal Chamber (Light)": 0xDA3BA362,
    "Putrid Alcove": 0xE212D644,
    "Ruined Alcove": 0xFAA9DF29,
    "Sacrificial Chamber": 0x4B426BC9,
    "Sacrificial Chamber Tunnel": 0x4C308AF0,
    "Save Station 1": 0xB3A28F7F,
    "Save Station 2": 0x3AF0B5FB,
    "Save Station A": 0xF2BBBB48,
    "Save Station B": 0xE2C3A0E5,
    "Temple Access": 0x2E954FEE,
    "Temple Transport Access": 0x0E0733A5,
    "Torvus Energy Controller": 0x3AEE0E85,
    "Torvus Grove": 0xC1ED0125,
    "Torvus Lagoon": 0x57DE1DA4,
    "Torvus Map Station": 0x01FEA342,
    "Torvus Plaza": 0x5C24A3C2,
    "Torvus Temple": 0xCF9535AB,
    "Training Access": 0x5E7A64EC,
    "Training Chamber": 0x6260555D,
    "Transit Tunnel East": 0x4F23CB33,
    "Transit Tunnel South": 0x9EF0619B,
    "Transit Tunnel West": 0xC8850070,
    "Transport to Agon Wastes": 0xE6B06473,
    "Transport to Sanctuary Fortress": 0x96DB1F15,
    "Transport to Temple Grounds": 0x46B0EECF,
    "Underground Transport": 0xBB8D27BF,
    "Underground Tunnel": 0xA93384AD,
    "Undertemple": 0xAC767B22,
    "Undertemple Access": 0x006C6BC5,
    "Undertemple Shaft": 0x9B061951,
    "Undertransit One": 0x750139C4,
    "Undertransit Two": 0x62295D04,
    "Venomous Pond": 0xAC5B17EE,
}
# Generated by asset_id_files.py

ABANDONED_WORKSITE_INTERNAL_ID = 0x8F9D66C0
AMMO_STATION_INTERNAL_ID = 0xA0E73B00
BROODING_GROUND_INTERNAL_ID = 0x3C156DDF
CACHE_A_INTERNAL_ID = 0x97B812A2
CACHE_B_INTERNAL_ID = 0xF04E9059
CATACOMBS_INTERNAL_ID = 0xAD9D5D9F
CATACOMBS_ACCESS_INTERNAL_ID = 0xC4F8E78E
CONTROLLER_ACCESS_INTERNAL_ID = 0x94182918
CRYPT_INTERNAL_ID = 0x4280A031
CRYPT_TUNNEL_INTERNAL_ID = 0xA29D33C8
DARK_ARENA_TUNNEL_INTERNAL_ID = 0xA1ECA232
DARK_CONTROLLER_ACCESS_INTERNAL_ID = 0x65B4E891
DARK_FALLS_INTERNAL_ID = 0x9172B3EB
DARK_FORGOTTEN_BRIDGE_INTERNAL_ID = 0xFF42758B
DARK_TORVUS_ARENA_INTERNAL_ID = 0xF5F50ABC
DARK_TORVUS_ENERGY_CONTROLLER_INTERNAL_ID = 0x8E2AF917
DARK_TORVUS_TEMPLE_INTERNAL_ID = 0x915E47CE
DARK_TORVUS_TEMPLE_ACCESS_INTERNAL_ID = 0x6DB75FB4
DUNGEON_INTERNAL_ID = 0xFE8931FC
FORGOTTEN_BRIDGE_INTERNAL_ID = 0x2406DD54
FORTRESS_TRANSPORT_ACCESS_INTERNAL_ID = 0x518D3511
GATHERING_ACCESS_INTERNAL_ID = 0x3AD40963
GATHERING_HALL_INTERNAL_ID = 0xF9F219B0
GLOOM_VISTA_INTERNAL_ID = 0x279818B4
GREAT_BRIDGE_INTERNAL_ID = 0x069D4087
GROVE_ACCESS_INTERNAL_ID = 0x734255B0
HYDROCHAMBER_STORAGE_INTERNAL_ID = 0x0112E539
HYDRODYNAMO_SHAFT_INTERNAL_ID = 0x3CE4A4FB
HYDRODYNAMO_STATION_INTERNAL_ID = 0x2A621228
MAIN_ENERGY_CONTROLLER_INTERNAL_ID = 0xDE01532E
MAIN_HYDROCHAMBER_INTERNAL_ID = 0x50039A19
MEDITATION_VISTA_INTERNAL_ID = 0xA701265D
PATH_OF_ROOTS_INTERNAL_ID = 0xC746378D
PLAZA_ACCESS_INTERNAL_ID = 0xE14A2729
POISONED_BOG_INTERNAL_ID = 0xA7DA4FD1
POLLUTED_MIRE_INTERNAL_ID = 0x8D13B482
PORTAL_CHAMBER_DARK_INTERNAL_ID = 0x9F510909
PORTAL_CHAMBER_LIGHT_INTERNAL_ID = 0x554E4514
PUTRID_ALCOVE_INTERNAL_ID = 0x0C8B7C06
RUINED_ALCOVE_INTERNAL_ID = 0x1D951459
SACRIFICIAL_CHAMBER_INTERNAL_ID = 0xC1AE6ECF
SACRIFICIAL_CHAMBER_TUNNEL_INTERNAL_ID = 0x7FF3F046
SAVE_STATION_1_INTERNAL_ID = 0x59E57214
SAVE_STATION_2_INTERNAL_ID = 0x0D580FBE
SAVE_STATION_A_INTERNAL_ID = 0x964465B0
SAVE_STATION_B_INTERNAL_ID = 0xA74BE553
TEMPLE_ACCESS_INTERNAL_ID = 0xE2F4F72A
TEMPLE_TRANSPORT_ACCESS_INTERNAL_ID = 0xAA2FA667
TORVUS_ENERGY_CONTROLLER_INTERNAL_ID = 0x3B696A7A
TORVUS_GROVE_INTERNAL_ID = 0x72349FA8
TORVUS_LAGOON_INTERNAL_ID = 0x0678EB72
TORVUS_MAP_STATION_INTERNAL_ID = 0xC1EC09C2
TORVUS_PLAZA_INTERNAL_ID = 0x1127C9E6
TORVUS_TEMPLE_INTERNAL_ID = 0x9A2ACAFD
TRAINING_ACCESS_INTERNAL_ID = 0x92BE0B06
TRAINING_CHAMBER_INTERNAL_ID = 0x1495E65B
TRANSIT_TUNNEL_EAST_INTERNAL_ID = 0x3B9904FD
TRANSIT_TUNNEL_SOUTH_INTERNAL_ID = 0xAEECD662
TRANSIT_TUNNEL_WEST_INTERNAL_ID = 0xA9917664
TRANSPORT_TO_AGON_WASTES_INTERNAL_ID = 0xA1197E74
TRANSPORT_TO_SANCTUARY_FORTRESS_INTERNAL_ID = 0x1C4BDD72
TRANSPORT_TO_TEMPLE_GROUNDS_INTERNAL_ID = 0x8D385440
UNDERGROUND_TRANSPORT_INTERNAL_ID = 0x8C23B6C3
UNDERGROUND_TUNNEL_INTERNAL_ID = 0x70FC85B3
UNDERTEMPLE_INTERNAL_ID = 0x22690759
UNDERTEMPLE_ACCESS_INTERNAL_ID = 0x416E3BA9
UNDERTEMPLE_SHAFT_INTERNAL_ID = 0x4DC6C3E3
UNDERTRANSIT_ONE_INTERNAL_ID = 0x3236D73D
UNDERTRANSIT_TWO_INTERNAL_ID = 0x02A8C6E4
VENOMOUS_POND_INTERNAL_ID = 0xAF3C03B1

NAME_TO_ID_INTERNAL_ID = {
    "Abandoned Worksite": 0x8F9D66C0,
    "Ammo Station": 0xA0E73B00,
    "Brooding Ground": 0x3C156DDF,
    "Cache A": 0x97B812A2,
    "Cache B": 0xF04E9059,
    "Catacombs": 0xAD9D5D9F,
    "Catacombs Access": 0xC4F8E78E,
    "Controller Access": 0x94182918,
    "Crypt": 0x4280A031,
    "Crypt Tunnel": 0xA29D33C8,
    "Dark Arena Tunnel": 0xA1ECA232,
    "Dark Controller Access": 0x65B4E891,
    "Dark Falls": 0x9172B3EB,
    "Dark Forgotten Bridge": 0xFF42758B,
    "Dark Torvus Arena": 0xF5F50ABC,
    "Dark Torvus Energy Controller": 0x8E2AF917,
    "Dark Torvus Temple": 0x915E47CE,
    "Dark Torvus Temple Access": 0x6DB75FB4,
    "Dungeon": 0xFE8931FC,
    "Forgotten Bridge": 0x2406DD54,
    "Fortress Transport Access": 0x518D3511,
    "Gathering Access": 0x3AD40963,
    "Gathering Hall": 0xF9F219B0,
    "Gloom Vista": 0x279818B4,
    "Great Bridge": 0x069D4087,
    "Grove Access": 0x734255B0,
    "Hydrochamber Storage": 0x0112E539,
    "Hydrodynamo Shaft": 0x3CE4A4FB,
    "Hydrodynamo Station": 0x2A621228,
    "Main Energy Controller": 0xDE01532E,
    "Main Hydrochamber": 0x50039A19,
    "Meditation Vista": 0xA701265D,
    "Path of Roots": 0xC746378D,
    "Plaza Access": 0xE14A2729,
    "Poisoned Bog": 0xA7DA4FD1,
    "Polluted Mire": 0x8D13B482,
    "Portal Chamber (Dark)": 0x9F510909,
    "Portal Chamber (Light)": 0x554E4514,
    "Putrid Alcove": 0x0C8B7C06,
    "Ruined Alcove": 0x1D951459,
    "Sacrificial Chamber": 0xC1AE6ECF,
    "Sacrificial Chamber Tunnel": 0x7FF3F046,
    "Save Station 1": 0x59E57214,
    "Save Station 2": 0x0D580FBE,
    "Save Station A": 0x964465B0,
    "Save Station B": 0xA74BE553,
    "Temple Access": 0xE2F4F72A,
    "Temple Transport Access": 0xAA2FA667,
    "Torvus Energy Controller": 0x3B696A7A,
    "Torvus Grove": 0x72349FA8,
    "Torvus Lagoon": 0x0678EB72,
    "Torvus Map Station": 0xC1EC09C2,
    "Torvus Plaza": 0x1127C9E6,
    "Torvus Temple": 0x9A2ACAFD,
    "Training Access": 0x92BE0B06,
    "Training Chamber": 0x1495E65B,
    "Transit Tunnel East": 0x3B9904FD,
    "Transit Tunnel South": 0xAEECD662,
    "Transit Tunnel West": 0xA9917664,
    "Transport to Agon Wastes": 0xA1197E74,
    "Transport to Sanctuary Fortress": 0x1C4BDD72,
    "Transport to Temple Grounds": 0x8D385440,
    "Underground Transport": 0x8C23B6C3,
    "Underground Tunnel": 0x70FC85B3,
    "Undertemple": 0x22690759,
    "Undertemple Access": 0x416E3BA9,
    "Undertemple Shaft": 0x4DC6C3E3,
    "Undertransit One": 0x3236D73D,
    "Undertransit Two": 0x02A8C6E4,
    "Venomous Pond": 0xAF3C03B1,
}

DOCK_NAMES = {
    "Abandoned Worksite": {
        "North": 0,
        "South": 1,
    },
    "Ammo Station": {
        "North": 0,
    },
    "Brooding Ground": {
        "North": 0,
        "South": 1,
    },
    "Cache A": {
        "North": 0,
        "VirtualNorth1": 1,
    },
    "Cache B": {
        "East_04": 0,
        "East_07": 1,
    },
    "Catacombs": {
        "North": 0,
        "West": 1,
        "West (West)": 2,
        "VirtualSouth": 3,
        "VirtualNorth1": 4,
    },
    "Catacombs Access": {
        "East": 0,
        "West": 1,
    },
    "Controller Access": {
        "docks_docks_West": 0,
        "East1": 1,
    },
    "Crypt": {
        "East_A": 0,
        "East_B": 1,
        "VirtualNorth": 2,
        "North": 3,
    },
    "Crypt Tunnel": {
        "East": 0,
        "West": 1,
    },
    "Dark Arena Tunnel": {
        "North": 0,
        "South": 1,
    },
    "Dark Controller Access": {
        "docks_docks_West": 0,
        "East1": 1,
    },
    "Dark Falls": {
        "West": 0,
        "East": 1,
    },
    "Dark Forgotten Bridge": {
        "North": 0,
        "East": 1,
        "South": 2,
        "VirtualNorth": 3,
        "West": 4,
    },
    "Dark Torvus Arena": {
        "South": 0,
    },
    "Dark Torvus Energy Controller": {
        "group6_East": 0,
        "North1": 1,
    },
    "Dark Torvus Temple": {
        "NorthDoor": 0,
        "South": 1,
        "West": 2,
        "WestGenerator": 3,
        "EastTop": 4,
        "East": 5,
    },
    "Dark Torvus Temple Access": {
        "Easttop": 0,
        "Eastbottom": 1,
        "Westtop": 2,
        "West": 3,
    },
    "Dungeon": {
        "North": 0,
        "West (mid)": 1,
        "West (West)": 2,
        "VirtualSouth": 3,
        "VirtualNorth1": 4,
    },
    "Forgotten Bridge": {
        "North": 0,
        "East": 1,
        "South": 2,
        "VirtualNorth": 3,
        "West": 4,
    },
    "Fortress Transport Access": {
        "South": 0,
        "North": 1,
    },
    "Gathering Access": {
        "East": 0,
        "West": 1,
    },
    "Gathering Hall": {
        "East_0L": 0,
        "East_0P": 1,
        "VirtualNorth": 2,
        "North": 3,
    },
    "Gloom Vista": {
        "Docks_South": 0,
        "VirtualNorth": 1,
    },
    "Great Bridge": {
        "WestUpper": 0,
        "WestLower": 1,
        "North": 2,
        "EastC": 3,
        "EastB": 4,
        "Docks_SouthUpper1": 5,
    },
    "Grove Access": {
        "West": 0,
        "East": 1,
    },
    "Hydrochamber Storage": {
        "North": 0,
    },
    "Hydrodynamo Shaft": {
        "Docks_VirtualNorth": 0,
        "Docks_North": 1,
        "West": 2,
    },
    "Hydrodynamo Station": {
        "West_0J": 0,
        "West_0L": 1,
        "North_OK": 2,
        "East_0M": 3,
        "East_Save": 4,
        "VirtualNorth": 5,
        "East (floor to 0q)": 6,
    },
    "Main Energy Controller": {
        "SouthLight1": 0,
        "VirtualNorth": 1,
        "North1": 2,
    },
    "Main Hydrochamber": {
        "South": 0,
        "VirtualNorth": 1,
        "South2": 2,
    },
    "Meditation Vista": {
        "Docks_South": 0,
        "VirtualNorth": 1,
    },
    "Path of Roots": {
        "West": 0,
        "North": 1,
    },
    "Plaza Access": {
        "North": 0,
        "South": 1,
    },
    "Poisoned Bog": {
        "NorthUpper": 0,
        "West": 1,
        "SouthLower": 2,
        "South": 3,
        "East": 4,
    },
    "Polluted Mire": {
        "East": 0,
        "South": 1,
        "North": 2,
    },
    "Portal Chamber (Dark)": {
        "East": 0,
        "VirtualNorth": 1,
        "West": 2,
    },
    "Portal Chamber (Light)": {
        "East": 0,
        "VirtualNorth": 1,
        "West": 2,
    },
    "Putrid Alcove": {
        "West": 0,
        "South": 1,
    },
    "Ruined Alcove": {
        "West": 0,
        "South": 1,
    },
    "Sacrificial Chamber": {
        "South": 0,
        "East": 1,
        "West1": 2,
        "North": 3,
    },
    "Sacrificial Chamber Tunnel": {
        "North": 0,
        "South": 1,
    },
    "Save Station 1": {
        "Docks_Docks_North": 0,
    },
    "Save Station 2": {
        "Docks_Docks_Docks_West": 0,
    },
    "Save Station A": {
        "North": 0,
        "VirtualNorth1": 1,
    },
    "Save Station B": {
        "Docks_Docks_Docks_West": 0,
    },
    "Temple Access": {
        "Easttop": 0,
        "Eastbottom": 1,
        "Westtop": 2,
        "West": 3,
    },
    "Temple Transport Access": {
        "East": 0,
        "West": 1,
    },
    "Torvus Energy Controller": {
        "group6_East": 0,
        "North1": 1,
    },
    "Torvus Grove": {
        "East": 0,
        "South": 1,
        "North": 2,
    },
    "Torvus Lagoon": {
        "NorthUpper (0D_Hall)": 0,
        "West (0C_Hall)": 1,
        "SouthLower (0B_Hall)": 2,
        "South (00_Pickup)": 3,
        "East (0A_Hall)": 4,
    },
    "Torvus Map Station": {
        "Docks_Docks_North": 0,
    },
    "Torvus Plaza": {
        "South": 0,
    },
    "Torvus Temple": {
        "NorthDoor": 0,
        "South": 1,
        "West": 2,
        "WestGenerator": 3,
        "EastTop": 4,
        "East": 5,
    },
    "Training Access": {
        "North": 0,
        "South": 1,
    },
    "Training Chamber": {
        "South": 0,
        "East": 1,
        "West1": 2,
        "North": 3,
    },
    "Transit Tunnel East": {
        "West": 0,
        "South": 1,
    },
    "Transit Tunnel South": {
        "East": 0,
        "West": 1,
    },
    "Transit Tunnel West": {
        "South": 0,
        "East": 1,
    },
    "Transport to Agon Wastes": {
        "North": 0,
    },
    "Transport to Sanctuary Fortress": {
        "South": 0,
    },
    "Transport to Temple Grounds": {
        "West": 0,
    },
    "Underground Transport": {
        "East_04 (Top)": 0,
        "East_07 (Bottom)": 1,
    },
    "Underground Tunnel": {
        "South": 0,
        "North": 1,
    },
    "Undertemple": {
        "South": 0,
        "VirtualNorth": 1,
        "South2": 2,
    },
    "Undertemple Access": {
        "Docks_VirtualNorth": 0,
        "Docks_North": 1,
        "West": 2,
    },
    "Undertemple Shaft": {
        "West_0J": 0,
        "West_0L": 1,
        "North_OK": 2,
        "East_0M": 3,
        "Docks_Docks_East_Save": 4,
        "VirtualNorth": 5,
        "East_0M1 (Floor to 0q)": 6,
    },
    "Undertransit One": {
        "South": 0,
        "East": 1,
    },
    "Undertransit Two": {
        "West": 0,
        "South": 1,
    },
    "Venomous Pond": {
        "WestUpper": 0,
        "WestLower": 1,
        "North": 2,
        "EastC": 3,
        "EastB": 4,
        "Docks_SouthUpper1": 5,
    },
}



import json
import os
from pathlib import Path

import pytest
from retro_data_structures.asset_manager import IsoFileProvider
from retro_data_structures.game_check import Game

from open_prime_rando.patcher_editor import PatcherEditor

_FAIL_INSTEAD_OF_SKIP = True


def get_env_or_skip(env_name: str, override_fail: bool | None = None) -> str:
    if override_fail is None:
        fail_or_skip = _FAIL_INSTEAD_OF_SKIP
    else:
        fail_or_skip = override_fail
    if env_name not in os.environ:
        if fail_or_skip:
            pytest.fail(f"Missing environment variable {env_name}")
        else:
            pytest.skip(f"Skipped due to missing environment variable {env_name}")
    return os.environ[env_name]


class TestFilesDir:
    def __init__(self, root: Path):
        self.root = root

    def joinpath(self, *paths) -> Path:
        return self.root.joinpath(*paths)

    def read_json(self, *paths) -> dict | list:
        with self.joinpath(*paths).open() as f:
            return json.load(f)


@pytest.fixture(scope="session")
def test_files_dir() -> TestFilesDir:
    return TestFilesDir(Path(__file__).parent.joinpath("test_files"))


@pytest.fixture(scope="module")
def prime2_ntsc_iso_path() -> Path:
    return Path(get_env_or_skip("PRIME2_ISO"))


@pytest.fixture(scope="module")
def prime2_pal_iso_path() -> Path:
    return Path(get_env_or_skip("PRIME2_PAL_ISO", override_fail=False))


@pytest.fixture(scope="module")
def prime2_iso_provider(prime2_ntsc_iso_path) -> IsoFileProvider:
    return IsoFileProvider(prime2_ntsc_iso_path)


@pytest.fixture(scope="module")
def pal_prime2_iso_provider(prime2_pal_iso_path) -> IsoFileProvider:
    return IsoFileProvider(prime2_pal_iso_path)


@pytest.fixture(scope="module")
def raw_prime2_editor(prime2_iso_provider: IsoFileProvider) -> PatcherEditor:
    return PatcherEditor(prime2_iso_provider, target_game=Game.ECHOES)


@pytest.fixture
def prime2_editor(raw_prime2_editor: PatcherEditor):
    editor = raw_prime2_editor
    yield editor
    editor._memory_files = {}
    for custom_asset, asset_id in editor._custom_asset_ids.items():
        editor._paks_for_asset_id.pop(asset_id)
    editor._custom_asset_ids = {}
    editor._modified_resources = {}


def pytest_addoption(parser) -> None:
    parser.addoption(
        "--skip-if-missing",
        action="store_false",
        dest="fail_if_missing",
        default=True,
        help="Skip tests instead of missing, in case any asset is missing",
    )


def pytest_configure(config: pytest.Config) -> None:
    global _FAIL_INSTEAD_OF_SKIP  # noqa: PLW0603
    _FAIL_INSTEAD_OF_SKIP = config.option.fail_if_missing



import logging
from collections.abc import Iterable

from retro_data_structures.enums.echoes import Message, State
from retro_data_structures.formats.mrea import Area
from retro_data_structures.formats.script_object import InstanceIdRef
from retro_data_structures.properties.echoes.archetypes.EditorProperties import EditorProperties
from retro_data_structures.properties.echoes.archetypes.LayerSwitch import LayerSwitch
from retro_data_structures.properties.echoes.archetypes.Transform import Transform
from retro_data_structures.properties.echoes.core.Vector import Vector
from retro_data_structures.properties.echoes.objects import (
    HUDMemo,
    Pickup,
    Relay,
    ScriptLayerController,
    SpawnPoint,
    Switch,
    Timer,
    Trigger,
)

from open_prime_rando.echoes.asset_ids import agon_wastes, sanctuary_fortress, temple_grounds, torvus_bog
from open_prime_rando.echoes.asset_ids.agon_wastes import COMMAND_CENTER_MREA
from open_prime_rando.echoes.asset_ids.world import (
    AGON_WASTES_MLVL,
    SANCTUARY_FORTRESS_MLVL,
    TEMPLE_GROUNDS_MLVL,
    TORVUS_BOG_MLVL,
)
from open_prime_rando.patcher_editor import PatcherEditor

LOG = logging.getLogger("echoes_patcher")


def apply_all(editor: PatcherEditor):
    """
    Applies changes necessary for the game to function properly.
    """
    mining_station_b(editor)
    undertemple_access(editor)
    main_reactor(editor)
    sacrificial_chamber(editor)
    aerie(editor)
    main_research(editor)
    hive_chamber_b(editor)
    gfmc_compound(editor)
    torvus_temple(editor)
    command_center_door(editor)


def mining_station_b(editor: PatcherEditor):
    """
    Normally this relay is activated by the cutscene, but has an incoming connection from the pickup.
    Activating it like this means the pickup will trigger the relay without the cutscene
    """
    area = editor.get_area(AGON_WASTES_MLVL, agon_wastes.MINING_STATION_B_MREA)
    with area.get_instance(0x80121).edit_properties(Relay) as post_pickup_relay:
        post_pickup_relay.editor_properties.active = True


def undertemple_access(editor: PatcherEditor):
    """
    Move the default spawn point in-bounds.
    """
    area = editor.get_area(TORVUS_BOG_MLVL, torvus_bog.UNDERTEMPLE_ACCESS_MREA)

    with area.get_instance("Spawn point 001").edit_properties(SpawnPoint) as spawn:
        spawn.editor_properties.transform.position.x = -149.25


def main_reactor(editor: PatcherEditor):
    """
    Save some memory during DS1 fight.
    """
    area = editor.get_area(AGON_WASTES_MLVL, agon_wastes.MAIN_REACTOR_MREA)

    unload_relay = area.get_instance("Unload dock when door closed")
    trigger = area.get_instance("Trigger Start DS Intro")
    trigger.add_connection(State.Entered, Message.SetToZero, unload_relay)


def sacrificial_chamber(editor: PatcherEditor):
    """
    Makes the pickup persistent, even if you exit the area and reload.
    """
    area = editor.get_area(TORVUS_BOG_MLVL, torvus_bog.SACRIFICIAL_CHAMBER_MREA)

    with area.get_instance("If grapple attainment is loaded, then end battle").edit_properties(Switch) as switch:
        switch.is_open = True


def _patch_echo_gate_softlock(
    area: Area, counter: InstanceIdRef, relays: Iterable[tuple[InstanceIdRef, InstanceIdRef]]
):
    for shot_relay_id, memory_relay_id in relays:
        shot_relay = area.get_instance(shot_relay_id)
        memory_relay = area.get_instance(memory_relay_id)

        shot_relay.remove_connections_from(counter)
        memory_relay.add_connection(State.Active, Message.Increment, counter)


def aerie(editor: PatcherEditor):
    """
    Patches an echo gate softlock.
    """
    area = editor.get_area(SANCTUARY_FORTRESS_MLVL, sanctuary_fortress.AERIE_MREA)

    relays = ((0x410094, 0x41008D), (0x410077, 0x41007F), (0x4100B5, 0x4100B6))
    _patch_echo_gate_softlock(area, 0x4100BE, relays)


def main_research(editor: PatcherEditor):
    """
    Patches an echo gate softlock.
    Disables the Contraption layer by default, activating it dynamically when exiting the lower portal.
    This hopefully prevents OOM/object list full crashes.
    """
    area = editor.get_area(SANCTUARY_FORTRESS_MLVL, sanctuary_fortress.MAIN_RESEARCH_MREA)

    relays = ((0x0B02E6, 0x0B02DE), (0x0B0303, 0x0B030D), (0x0B02F6, 0x0B02FA))
    _patch_echo_gate_softlock(area, 0x0B0315, relays)

    area.get_layer("Contraption").active = False
    portal_spawn = area.get_instance(0x0B0056)
    spawn_xfm = portal_spawn.get_properties_as(SpawnPoint).editor_properties.transform

    contraption_controller = area.get_layer("Default").add_instance_with(
        ScriptLayerController(
            editor_properties=EditorProperties(name="Dynamic Increment Contraption", transform=spawn_xfm),
            layer=LayerSwitch(
                area_id=sanctuary_fortress.MAIN_RESEARCH_INTERNAL_ID, layer_number=area.get_layer("Contraption").index
            ),
            is_dynamic=True,
        )
    )
    spider_controller = area.get_layer("Default").add_instance_with(
        ScriptLayerController(
            editor_properties=EditorProperties(name="Dynamic Increment Column Spiderball", transform=spawn_xfm),
            layer=LayerSwitch(
                area_id=sanctuary_fortress.MAIN_RESEARCH_INTERNAL_ID,
                layer_number=area.get_layer("Column Spiderball").index,
            ),
            is_dynamic=True,
        )
    )

    for controller in (contraption_controller, spider_controller):
        portal_spawn.add_connection(State.Arrived, Message.Load, controller)
        controller.add_connection(State.Arrived, Message.Play, controller)


def hive_chamber_b(editor: PatcherEditor):
    """
    Removes item loss sequence.
    """
    area = editor.get_area(TEMPLE_GROUNDS_MLVL, temple_grounds.HIVE_CHAMBER_B_MREA)

    # FIXME: use layers API
    area.get_layer("DS Appears Part1").active = False
    area.get_layer("Pre Dark Samus Music").active = False

    area.get_layer("Pickup").active = True
    area.get_layer("Post Dark Samus Music").active = True


def gfmc_compound(editor: PatcherEditor):
    """
    Add a HUDMemo for the ship missile.
    """
    area = editor.get_area(TEMPLE_GROUNDS_MLVL, temple_grounds.GFMC_COMPOUND_MREA)

    pickup_xfm = area.get_instance(0x2B0324).get_properties_as(Pickup).editor_properties.transform
    ship_trigger = area.get_layer("Default").add_instance_with(
        Trigger(
            editor_properties=EditorProperties(
                name="Show Ship Missile HudMemo",
                transform=Transform(
                    position=pickup_xfm.position, rotation=Vector(0.0, 0.0, 45.0), scale=Vector(50.0, 50.0, 10.0)
                ),
            ),
            deactivate_on_enter=True,
        )
    )
    strg_id, _ = editor.create_strg(
        "gfmc_jump_hudmemo.STRG", ["Defeating Jump Guardian is required for this item to appear."]
    )
    hud_memo = area.get_layer("Default").add_instance_with(
        HUDMemo(
            editor_properties=EditorProperties(name="Ship Missile HudMemo", transform=pickup_xfm),
            display_time=6.0,
            display_type=0,
            string=strg_id,
        )
    )
    ship_trigger.add_connection(State.Entered, Message.SetToZero, hud_memo)

    timer = area.get_layer("Space Jump").add_instance_with(
        Timer(
            editor_properties=EditorProperties(name="Disable Ship Missile Trigger", transform=pickup_xfm),
            time=0.1,
            auto_start=True,
        )
    )
    timer.add_connection(State.Zero, Message.Deactivate, ship_trigger)


def torvus_temple(editor: PatcherEditor):
    """
    Remove cosmetic objects from Torvus Temple to minimize the chance of crash via alloc failure
    """
    area = editor.get_area(TORVUS_BOG_MLVL, torvus_bog.TORVUS_TEMPLE_MREA)

    to_remove = [
        "Thrust1",
        "Thrust1",
        "Thrust2",
        "Thrust2",
        "Looping Thrust w/Doppler",
        "Looping Thrust w/Doppler",
        "GENERATE GIBS",
    ]
    to_remove.extend(["SwampCrateDebris"] * 7)

    for obj in to_remove:
        area.remove_instance(obj)


def command_center_door(editor: PatcherEditor):
    """
    Opening the blast door normally requires a room reload after they've been closed.
    The DS cutscene in Security Station B reloads the room, but that cutscene has been removed.
    """
    area = editor.get_area(AGON_WASTES_MLVL, COMMAND_CENTER_MREA)
    default = area.get_layer("Default")

    poi = default.get_instance("Blast Door Activation")

    # Deactivate layers
    for layer in ("1st Pass Scripting", "1st pass parts"):
        controller = default.add_instance_with(
            ScriptLayerController(
                editor_properties=EditorProperties(
                    name=f"DYNAMIC Decrement {layer}",
                ),
                layer=LayerSwitch(
                    area_id=0xAA657163,
                    layer_number=area.get_layer(layer).index,
                ),
                is_dynamic=True,
            )
        )
        poi.add_connection(State.ScanDone, Message.Decrement, controller)

    # Ensure the blast door instances are active for the cutscene
    for instance in ("Upper Blast Door", "Lower Blast Door"):
        poi.add_connection(
            State.ScanDone,
            Message.Activate,
            default.get_instance(instance),
        )



from retro_data_structures.enums.echoes import Message, State
from retro_data_structures.properties.echoes.archetypes.EditorProperties import EditorProperties
from retro_data_structures.properties.echoes.objects.Timer import Timer

from open_prime_rando.echoes.asset_ids import (
    agon_wastes,
    great_temple,
    sanctuary_fortress,
    temple_grounds,
    torvus_bog,
    world,
)
from open_prime_rando.patcher_editor import PatcherEditor

ELEVATOR_MEMORY_RELAY_PER_MREA = {
    world.GREAT_TEMPLE_MLVL: {
        great_temple.TEMPLE_TRANSPORT_A_MREA: 0x00000107,
        great_temple.TEMPLE_TRANSPORT_B_MREA: 0x0008002F,
        great_temple.TEMPLE_TRANSPORT_C_MREA: 0x00060046,
    },
    world.TEMPLE_GROUNDS_MLVL: {
        temple_grounds.TRANSPORT_TO_AGON_WASTES_MREA: 0x00180040,
        temple_grounds.TRANSPORT_TO_SANCTUARY_FORTRESS_MREA: 0x00330058,
        temple_grounds.TRANSPORT_TO_TORVUS_BOG_MREA: 0x001E0061,
    },
    world.AGON_WASTES_MLVL: {
        agon_wastes.TRANSPORT_TO_SANCTUARY_FORTRESS_MREA: 0x002D0064,
        agon_wastes.TRANSPORT_TO_TEMPLE_GROUNDS_MREA: 0x0000003F,
        agon_wastes.TRANSPORT_TO_TORVUS_BOG_MREA: 0x0013004D,
    },
    world.TORVUS_BOG_MLVL: {
        torvus_bog.TRANSPORT_TO_AGON_WASTES_MREA: 0x00210038,
        torvus_bog.TRANSPORT_TO_SANCTUARY_FORTRESS_MREA: 0x0045004B,
        torvus_bog.TRANSPORT_TO_TEMPLE_GROUNDS_MREA: 0x0000002E,
    },
    world.SANCTUARY_FORTRESS_MLVL: {
        sanctuary_fortress.TRANSPORT_TO_AGON_WASTES_MREA: 0x00130088,
        sanctuary_fortress.TRANSPORT_TO_TEMPLE_GROUNDS_MREA: 0x00000035,
        sanctuary_fortress.TRANSPORT_TO_TORVUS_BOG_MREA: 0x00190036,
    },
}


def apply_auto_enabled_elevators_patch(editor: PatcherEditor):
    """
    Patches that activates every elevator on room load
    """
    for mlvl_id, areas in ELEVATOR_MEMORY_RELAY_PER_MREA.items():
        for mrea_id, memory_relay_id in areas.items():
            area = editor.get_area(mlvl_id, mrea_id)
            timer = area.get_layer("Default").add_instance_with(
                Timer(
                    editor_properties=EditorProperties(name="Timer - Auto enable elevator", active=False),
                    time=0.001,
                    auto_start=True,
                )
            )
            relay = area.get_instance(memory_relay_id)
            timer.add_connection(State.Zero, Message.Activate, relay)



from retro_data_structures.formats import Mapa, Mlvl
from retro_data_structures.formats.mrea import Area
from retro_data_structures.formats.script_layer import ScriptLayer
from retro_data_structures.formats.script_object import InstanceId, ScriptInstance
from retro_data_structures.properties.echoes.objects.AreaAttributes import AreaAttributes
from retro_data_structures.properties.echoes.objects.SafeZone import SafeZone
from retro_data_structures.properties.echoes.objects.SafeZoneCrystal import SafeZoneCrystal

from open_prime_rando.echoes.asset_ids import world as world_ids
from open_prime_rando.echoes.inverted import area_pairs
from open_prime_rando.patcher_editor import PatcherEditor

_WORLDS = [
    world_ids.TEMPLE_GROUNDS_MLVL,
    world_ids.AGON_WASTES_MLVL,
    world_ids.TORVUS_BOG_MLVL,
    world_ids.SANCTUARY_FORTRESS_MLVL,
    world_ids.GREAT_TEMPLE_MLVL,
]
_AREAS_TO_SKIP = {
    0x775664A5,  # 00_scandummy
    0x752B2850,  # game_end_part1
    0x9D221F4A,  # game_end_part2
    0xC5250DBC,  # game_end_part3
    0x9641773F,  # game_end_part4
    0xCE4665C9,  # game_end_part5
    0x4D4E5400,  # Menu
}


def _swap_dark_world(editor: PatcherEditor):
    for world_id in _WORLDS:
        world = editor.get_mlvl(world_id)

        for area in world.areas:
            if area.id in _AREAS_TO_SKIP:
                continue

            is_dark_world = area.internal_name.endswith("_dark")

            for instance in area.all_instances:
                if instance.type == AreaAttributes:
                    with instance.edit_properties(AreaAttributes) as prop:
                        if prop.dark_world != is_dark_world:
                            print(area.name, is_dark_world, "found", prop.dark_world)
                        prop.dark_world = not is_dark_world

            mapa_id = world.mapw.get_mapa_id(area.index)
            mapa = editor.get_file(mapa_id, Mapa)
            mapa.raw.type = not is_dark_world


def _copy_safe_zones(dark: Area, dark_layer: ScriptLayer, light_layer: ScriptLayer, special_ids: set):
    copied_safe_zones = {}

    for instance in list(dark_layer.instances):
        if instance.type == SafeZone:
            copied_safe_zones[instance.id] = light_layer.add_instance_with(instance.get_properties_as(SafeZone))

            ids = {conn.target for conn in instance.connections}
            for target in ids:
                target_name = dark.get_instance(target).name
                if "ENTERED Safezone" not in target_name and "EXITED Safezone" not in target_name:
                    special_ids.add(instance.id)

            if instance.id not in special_ids:
                dark_layer.remove_instance(instance)

    return copied_safe_zones


def _copy_safe_zone_crystal(
    copied_crystal: ScriptInstance,
    instance: ScriptInstance,
    copied_safe_zones: dict[InstanceId, ScriptInstance],
    dark: Area,
    special_ids: set[int],
):
    targets_special = False

    for conn in instance.connections:
        if conn.target == instance.id:
            target = copied_crystal
        elif conn.target in copied_safe_zones:
            targets_special = targets_special or conn.target in special_ids
            target = copied_safe_zones[conn.target]
        else:
            targets_special = True
            weird_target = dark.get_instance(conn.target)
            print(
                f"Crystal {instance.name} at {dark.name} has unexpected connections "
                f"to {weird_target} ({weird_target.name})"
            )
            continue
        copied_crystal.add_connection(conn.state, conn.message, target)

    return targets_special


def _move_safe_zones(world: Mlvl, pairs: list[tuple[int, int]]):
    for light_id, dark_id in pairs:
        light = world.get_area(light_id)
        dark = world.get_area(dark_id)

        light_layer = light.get_layer("Default")
        dark_layer = dark.get_layer("Default")

        assert light_layer.index == 0

        special_ids = set()  # ids for objects that have extra functionality and we want to keep

        # Copy the safe zones
        copied_safe_zones = _copy_safe_zones(dark, dark_layer, light_layer, special_ids)

        # Copy the safe zone crystals
        for instance in list(dark_layer.instances):
            if instance.type == SafeZoneCrystal:
                copied_crystal = light_layer.add_instance_with(instance.get_properties_as(SafeZoneCrystal))

                targets_special = _copy_safe_zone_crystal(
                    copied_crystal, instance, copied_safe_zones, dark, special_ids
                )
                if not targets_special:
                    dark_layer.remove_instance(instance)


def apply_inverted(editor: PatcherEditor):
    _swap_dark_world(editor)

    _move_safe_zones(editor.get_mlvl(world_ids.TEMPLE_GROUNDS_MLVL), area_pairs.TG_PAIRS)
    _move_safe_zones(editor.get_mlvl(world_ids.AGON_WASTES_MLVL), area_pairs.AGON_PAIRS)
    _move_safe_zones(editor.get_mlvl(world_ids.TORVUS_BOG_MLVL), area_pairs.TORVUS_PAIRS)
    _move_safe_zones(editor.get_mlvl(world_ids.SANCTUARY_FORTRESS_MLVL), area_pairs.SANCTUARY_PAIRS)
    _move_safe_zones(editor.get_mlvl(world_ids.GREAT_TEMPLE_MLVL), area_pairs.GREAT_TEMPLE_PAIRS)



import dataclasses
import functools
from pathlib import Path

from retro_data_structures.base_resource import AssetId, NameOrAssetId, RawResource
from retro_data_structures.formats import Ancs, Cmdl

from open_prime_rando.patcher_editor import PatcherEditor


@functools.cache
def custom_asset_path() -> Path:
    return Path(__file__).parent


SCAN_VISOR_CMDL = 0xAFC70004
COMBAT_VISOR_CMDL = 0x0B7E6CA9
DARK_VISOR_ANCS = 0x851B526E


@dataclasses.dataclass(frozen=True)
class TranslatorAssets:
    txtr: NameOrAssetId
    new_cmdl: int | None = None
    new_ancs: int | None = None


TRANSLATORS: dict[str, TranslatorAssets] = {
    "violet_translator": TranslatorAssets(
        txtr=0x4BE5342E,
        new_cmdl=1448365380,
        new_ancs=1448362318,
    ),
    "amber_translator": TranslatorAssets(
        txtr=0xF5308558,
        new_cmdl=1096043844,
        new_ancs=1096040782,
    ),
    "emerald_translator": TranslatorAssets(
        txtr=0xA9640FDF,
        new_cmdl=1163152708,
        new_ancs=1163149646,
    ),
    "cobalt_translator": TranslatorAssets(
        txtr=0x2C56D2D4,
        new_cmdl=1129598276,
        new_ancs=1129595214,
    ),
}


def _create_visor_derivatives(editor: PatcherEditor):
    cmdls: dict[str, AssetId] = {
        "combat_visor": COMBAT_VISOR_CMDL,
        "scan_visor": SCAN_VISOR_CMDL,
    }

    for name, assets in TRANSLATORS.items():
        if assets.new_ancs is not None and assets.new_cmdl is not None:
            editor.register_custom_asset_name(f"{name}.CMDL", assets.new_cmdl)
            editor.register_custom_asset_name(f"{name}.ANCS", assets.new_ancs)

        texture = editor._resolve_asset_id(assets.txtr)
        cmdl_id = editor.duplicate_asset(SCAN_VISOR_CMDL, f"{name}.CMDL")

        cmdl = editor.get_file(cmdl_id, type_hint=Cmdl)
        cmdl.raw.material_sets[0].texture_file_ids[0] = texture
        cmdl.raw.material_sets[0].texture_file_ids[1] = texture

        cmdls[name] = cmdl_id

    for name, model in cmdls.items():
        ancs_id = editor.duplicate_asset(DARK_VISOR_ANCS, f"{name}.ANCS")
        ancs = editor.get_file(ancs_id, type_hint=Ancs)
        ancs.raw.character_set.characters[0].model_id = model


@dataclasses.dataclass
class BeamAmmoAssets:
    txtr_a: int
    txtr_b: int
    particle: int
    new_cmdl: int
    new_ancs: int


BEAM_AMMO_EXPANSION_CMDL = 0x352C8B02
BEAM_AMMO_EXPANSION_ANCS = 0x4E00188C

BEAM_AMMO = {
    "dark_ammo": BeamAmmoAssets(
        txtr_a=1385637646,
        txtr_b=3277274054,
        particle=0x8E0C499D,
        new_cmdl=1631670273,
        new_ancs=0x61415002,
    ),
    "light_ammo": BeamAmmoAssets(
        txtr_a=1815959726,
        txtr_b=2738959665,
        particle=0xA180BB7F,
        new_cmdl=1631670275,
        new_ancs=1631670276,
    ),
}


def _create_split_ammo(editor: PatcherEditor):
    for name, ammo in BEAM_AMMO.items():
        editor.register_custom_asset_name(f"{name}.CMDL", ammo.new_cmdl)
        editor.register_custom_asset_name(f"{name}.ANCS", ammo.new_ancs)

        cmdl_id = editor.duplicate_asset(BEAM_AMMO_EXPANSION_CMDL, f"{name}.CMDL")
        cmdl = editor.get_file(cmdl_id, type_hint=Cmdl)
        cmdl.raw.material_sets[0].texture_file_ids[0] = ammo.txtr_a
        cmdl.raw.material_sets[0].texture_file_ids[2] = ammo.txtr_a
        cmdl.raw.material_sets[0].texture_file_ids[3] = ammo.txtr_b
        cmdl.raw.material_sets[0].texture_file_ids[4] = ammo.txtr_b

        ancs_id = editor.duplicate_asset(BEAM_AMMO_EXPANSION_ANCS, f"{name}.ANCS")
        ancs = editor.get_file(ancs_id, type_hint=Ancs)
        ancs.raw.character_set.characters[0].model_id = ammo.new_cmdl
        ancs.raw.character_set.characters[0].particle_resource_data.generic_particles = [
            0x89C2F268,
            ammo.particle,
        ]
        ancs.raw.animation_set.event_sets[0].particle_poi_nodes[0].particle.id = ammo.particle
        ancs.raw.animation_set.event_sets[0].particle_poi_nodes[1].particle.id = ammo.particle


def _import_premade_assets(editor: PatcherEditor):
    assets = custom_asset_path().joinpath("general")
    for f in assets.rglob("*.*"):
        name = f.name
        asset_type = f.suffix[1:].upper()  # remove leading period, force uppercase
        raw = f.read_bytes()
        editor.add_new_asset(name, RawResource(asset_type, raw), ())


def create_custom_assets(editor: PatcherEditor):
    _import_premade_assets(editor)

    _create_visor_derivatives(editor)
    _create_split_ammo(editor)



from typing import Annotated

from annotated_types import Interval
from open_prime_rando_practice_mod import PracticeModMode
from pydantic import BaseModel

AssetId = Annotated[int, Interval(ge=0, le=0xFFFFFFFF)]


class AreaReference(BaseModel):
    mlvl_id: AssetId
    mrea_id: AssetId


class RandoConfiguration(BaseModel):
    """
    Configuration for randomizing Metroid Prime 2: Echoes.
    """

    starting_area: AreaReference | None = None
    """The game will start at the given area. When not set, starts at Landing Site."""

    practice_mod: PracticeModMode = PracticeModMode.disabled
    """How accessible is the Practice Mod."""



import argparse
import collections
import os
import typing
from pathlib import Path

import graphviz
import networkx
from retro_data_structures.asset_manager import IsoFileProvider
from retro_data_structures.formats.script_object import ScriptInstance
from retro_data_structures.game_check import Game

from open_prime_rando.echoes.asset_ids import world
from open_prime_rando.patcher_editor import PatcherEditor


def _draw_component(
    dot: graphviz.Digraph, instances: list[ScriptInstance], namer: typing.Callable[[ScriptInstance], str]
):
    for instance in instances:
        dot.node(name=str(instance.id), label=namer(instance))

    for instance in instances:
        for con in instance.connections:
            dot.edge(str(instance.id), str(con.target), label=f"{con.state.name} - {con.message.name}")

    dot.render(cleanup=True, format="png")


def draw_objects(editor: PatcherEditor, world_name: str, area_name: str):
    mlvl = editor.get_mlvl(world.NAME_TO_ID_MLVL[world_name])
    area = mlvl.get_area(world.load_dedicated_file(world_name).NAME_TO_ID_MREA[area_name])

    graph = networkx.DiGraph()

    name_count: dict[str, int] = collections.defaultdict(int)

    for instance in area.all_instances:
        name_count[instance.name] += 1
        graph.add_node(instance.id)
        for con in instance.connections:
            graph.add_node(con.target)
            graph.add_edge(instance.id, con.target)

    def namer(inst: ScriptInstance):
        name = inst.name
        if name:
            if name_count[name] > 1:
                name += f" ({inst.id})"
        else:
            name = str(inst.id)
        return name

    group_i = 1
    for comp in networkx.weakly_connected_components(graph):
        if len(comp) < 2:
            continue

        _draw_component(
            graphviz.Digraph(name=f"{area_name} ({group_i})"),
            [instance for instance in area.all_instances if instance.id in comp],
            namer,
        )

        group_i += 1


def main():
    iso = os.getenv("PRIME2_ISO")

    parser = argparse.ArgumentParser()
    parser.add_argument("--game", required=False, default="echoes", choices=["echoes"])
    parser.add_argument("--iso", required=iso is None, type=Path, default=iso, help="Path to where the ISO.")
    parser.add_argument("world", help="Name of the world to check, as used in the schema")
    parser.add_argument("area", help="Name of the area to check, as used in the schema")
    args = parser.parse_args()

    draw_objects(
        PatcherEditor(IsoFileProvider(args.iso), Game.ECHOES),
        args.world,
        args.area,
    )


if __name__ == "__main__":
    main()



import dataclasses
import uuid

from ppc_asm.dol_file import DolEditor

from open_prime_rando.dol_patching import all_prime_dol_patches, dol_version
from open_prime_rando.dol_patching.echoes import dol_patches, dol_versions
from open_prime_rando.dol_patching.echoes.beam_configuration import BeamAmmoConfiguration
from open_prime_rando.dol_patching.echoes.user_preferences import OprEchoesUserPreferences


@dataclasses.dataclass(frozen=True)
class EchoesDolPatchesData:
    world_uuid: uuid.UUID
    energy_per_tank: int
    beam_configurations: list[BeamAmmoConfiguration]
    safe_zone_heal_per_second: float
    user_preferences: OprEchoesUserPreferences
    default_items: dict
    unvisited_room_names: bool
    teleporter_sounds: bool
    dangerous_energy_tank: bool

    @classmethod
    def from_json(cls, data: dict):
        return cls(
            world_uuid=uuid.UUID(data["world_uuid"]),
            energy_per_tank=data["energy_per_tank"],
            beam_configurations=[BeamAmmoConfiguration.from_json(it) for it in data["beam_configurations"]],
            safe_zone_heal_per_second=data["safe_zone_heal_per_second"],
            user_preferences=OprEchoesUserPreferences.from_json(data["user_preferences"]),
            default_items=data["default_items"],
            unvisited_room_names=data["unvisited_room_names"],
            teleporter_sounds=data["teleporter_sounds"],
            dangerous_energy_tank=data["dangerous_energy_tank"],
        )


def apply_patches(dol_editor: DolEditor, patches_data: EchoesDolPatchesData) -> dol_patches.EchoesDolVersion:
    version = dol_version.find_version_for_dol(dol_editor, dol_versions.ALL_VERSIONS)
    assert isinstance(version, dol_patches.EchoesDolVersion)

    dol_patches.apply_fixes(version, dol_editor)
    dol_patches.change_powerup_should_persist(
        version, dol_editor, ["Double Damage", "Unlimited Missiles", "Unlimited Beam Ammo"]
    )

    all_prime_dol_patches.apply_build_info_patch(dol_editor, patches_data.world_uuid, version)
    all_prime_dol_patches.apply_remote_execution_patch(version.game, version.string_display, dol_editor)
    all_prime_dol_patches.apply_energy_tank_capacity_patch(
        version.health_capacity, patches_data.energy_per_tank, dol_editor
    )
    all_prime_dol_patches.apply_reverse_energy_tank_heal_patch(
        version.sda2_base, version.dangerous_energy_tank, patches_data.dangerous_energy_tank, version.game, dol_editor
    )

    dol_patches.apply_unvisited_room_names(version, dol_editor, patches_data.unvisited_room_names)
    dol_patches.apply_teleporter_sounds(version, dol_editor, patches_data.teleporter_sounds)

    dol_patches.apply_game_options_patch(
        version.game_options_constructor_address, patches_data.user_preferences, dol_editor
    )
    dol_patches.apply_beam_cost_patch(version.beam_cost_addresses, patches_data.beam_configurations, dol_editor)
    dol_patches.apply_safe_zone_heal_patch(
        version.safe_zone, version.sda2_base, patches_data.safe_zone_heal_per_second, dol_editor
    )
    dol_patches.apply_starting_visor_patch(version.starting_beam_visor, patches_data.default_items, dol_editor)
    dol_patches.apply_map_door_changes(version.map_door_types, dol_editor)

    return version



from open_prime_rando.echoes.specific_area_patches import rebalance_patches, required_fixes
from open_prime_rando.echoes.specific_area_patches.version_differences import EchoesVersion, patch_version_differences
from open_prime_rando.patcher_editor import PatcherEditor


def specific_patches(editor: PatcherEditor, area_patches: dict):
    required_fixes.apply_all(editor)
    patch_version_differences(editor, EchoesVersion.NTSC_U)  # TODO: detect version
    if area_patches["rebalance_world"]:
        rebalance_patches.apply_all(editor)



import dataclasses
from collections.abc import Iterable

from ppc_asm.dol_file import DolEditor
from retro_data_structures.game_check import Game


@dataclasses.dataclass(frozen=True)
class DolVersion:
    game: Game
    description: str
    build_string_address: int
    build_string: bytes
    sda2_base: int
    sda13_base: int


def find_version_for_dol(dol_editor: DolEditor, all_versions: Iterable[DolVersion]) -> DolVersion:
    for version in all_versions:
        build_string = dol_editor.read(version.build_string_address, len(version.build_string))
        if (build_string[:6], build_string[6 + 16 :]) == (version.build_string[:6], version.build_string[6 + 16 :]):
            return version

    raise RuntimeError("Unsupported game version")



from retro_data_structures.formats import Strg

from open_prime_rando.patcher_editor import PatcherEditor


def apply_corrupted_memory_card_change(editor: PatcherEditor):
    # STRG_MemoryCard_0
    table = editor.get_file(0x88E242D6, Strg)

    table.set_single_string(
        table.raw.name_table["CorruptedFile"],
        """The save file was created using a different
Randomizer ISO and must be deleted.""",
    )
    table.set_single_string(table.raw.name_table["ChoiceDeleteCorruptedFile"], "Delete Incompatible File")



import dataclasses

from retro_data_structures.properties.echoes.archetypes.VisorParameters import VisorFlags
from retro_data_structures.properties.echoes.core.Color import Color
from retro_data_structures.properties.echoes.core.Vector import Vector

from open_prime_rando.echoes.dock_lock_rando import dock_type
from open_prime_rando.echoes.dock_lock_rando.map_icons import DoorMapIcon
from open_prime_rando.echoes.vulnerabilities import normal_vuln, resist_all_vuln, vulnerable, vulnerable_no_splash

normal_door_model = 0x6B78FD92
dark_door_model = 0xBBCF134D
annihilator_door_model = 0xA50C7238

blast_collision_box = Vector(0.35, 5.0, 4.0)
blast_collision_offset = Vector(-2 / 3, 0, 2.0)

DOCK_TYPES: dict[str, dock_type.DoorType] = {
    "Normal": dock_type.NormalDoorType(
        name="Normal",
        vulnerability=normal_vuln,
        shell_color=Color(0, 1, 1, 1),
        map_icon=DoorMapIcon.Normal,
    ),
    "Dark": dock_type.NormalDoorType(
        name="Dark",
        vulnerability=dataclasses.replace(
            resist_all_vuln, dark=vulnerable, entangler=vulnerable, black_hole=vulnerable
        ),
        shell_model=dark_door_model,
        shell_color=Color(1, 1, 1, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. Dark energy may damage it.",
        ),
        map_icon=DoorMapIcon.Dark,
    ),
    "Light": dock_type.NormalDoorType(
        name="Light",
        vulnerability=dataclasses.replace(
            resist_all_vuln, light=vulnerable, light_blast=vulnerable, sunburst=vulnerable
        ),
        shell_color=Color(1, 1, 1, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. Light energy may damage it.",
        ),
        map_icon=DoorMapIcon.Light,
    ),
    "Annihilator": dock_type.NormalDoorType(
        name="Annihilator",
        vulnerability=dataclasses.replace(
            resist_all_vuln, annihilator=vulnerable, sonic_boom=vulnerable, imploder=vulnerable
        ),
        shell_model=annihilator_door_model,
        shell_color=Color(1, 1, 1, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "A mix of light and dark energy may damage it.",
        ),
        map_icon=DoorMapIcon.Annihilator,
    ),
    "Disabled": dock_type.NormalDoorType(
        name="Disabled",
        vulnerability=resist_all_vuln,
        shell_color=Color(0, 0, 0, 0),
        scan_text=("Door system access denied.", "Unable to bypass security codes. Seek an alternate exit."),
        map_icon=DoorMapIcon.Disabled,
    ),
    "Missile": dock_type.VanillaBlastShieldDoorType(
        name="Missile",
        vulnerability=dataclasses.replace(resist_all_vuln, missile=vulnerable),
        shell_color=Color(1, 0, 0, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. A Missile blast may damage it.",
        ),
        map_icon=DoorMapIcon.Missile,
        shield_model=0xBFB4A8EE,
    ),
    "SuperMissile": dock_type.VanillaBlastShieldDoorType(
        name="SuperMissile",
        vulnerability=dataclasses.replace(resist_all_vuln, super_missle=vulnerable),
        shell_color=Color(0, 1, 0, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "A Super Missile blast may damage it.",
        ),
        map_icon=DoorMapIcon.SuperMissile,
        shield_model=0xF115F575,
    ),
    "PowerBomb": dock_type.VanillaBlastShieldDoorType(
        name="PowerBomb",
        vulnerability=dataclasses.replace(resist_all_vuln, power_bomb=vulnerable),
        shell_color=Color(1, 0.94, 0, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. A Power Bomb may damage it.",
        ),
        map_icon=DoorMapIcon.PowerBomb,
        shield_model=0xC2E4F075,
    ),
    "SeekerMissile": dock_type.SeekerBlastShieldDoorType(
        name="SeekerMissile",
        vulnerability=dataclasses.replace(resist_all_vuln, missile=vulnerable),
        shell_color=Color(0.5, 0, 0.64, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "Five simultaneous Missile blasts may damage it.",
        ),
        map_icon=DoorMapIcon.SeekerMissile,
        shield_model=0x56F4208B,
    ),
    "ScrewAttack": dock_type.BlastShieldDoorType(
        name="ScrewAttack",
        vulnerability=dataclasses.replace(resist_all_vuln, screw_attack=vulnerable),
        shell_model=normal_door_model,
        shell_color=Color(0.93, 0.58, 0.83, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. The Screw Attack may damage it.",
        ),
        map_icon=DoorMapIcon.ScrewAttack,
        shield_model="screw_attack",
    ),
    "Bomb": dock_type.BlastShieldDoorType(
        name="Bomb",
        vulnerability=dataclasses.replace(resist_all_vuln, bomb=vulnerable),
        shell_model=normal_door_model,
        shell_color=Color(1 / 3, 1 / 3, 0.5, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "A Morph Ball Bomb blast may damage it.",
        ),
        map_icon=DoorMapIcon.Bomb,
        shield_model="morph_ball_bombs",
    ),
    "Boost": dock_type.BlastShieldDoorType(
        name="Boost",
        vulnerability=dataclasses.replace(resist_all_vuln, boost_ball=vulnerable, cannon_ball=vulnerable_no_splash),
        shell_model=normal_door_model,
        shell_color=Color(1, 1 / 3, 0, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. The Boost Ball may damage it.",
        ),
        map_icon=DoorMapIcon.Boost,
        shield_model="boost_ball",
    ),
    "Grapple": dock_type.GrappleDoorType(
        name="Grapple",
        vulnerability=resist_all_vuln,
        shell_model=annihilator_door_model,
        shell_color=Color(1, 0, 0, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to weapon fire, but is weakly secured. "
            "The Grapple Beam may be able to remove it.",
        ),
        map_icon=DoorMapIcon.Grapple,
        shield_model=0x56F4208B,
    ),
    "Darkburst": dock_type.BlastShieldDoorType(
        name="Darkburst",
        vulnerability=dataclasses.replace(resist_all_vuln, black_hole=vulnerable),
        shell_model=dark_door_model,
        shell_color=Color(1, 1, 1, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "A massive burst of dark energy may damage it.",
        ),
        map_icon=DoorMapIcon.Dark,
        shield_model="darkburst",
    ),
    "Sunburst": dock_type.BlastShieldDoorType(
        name="Sunburst",
        vulnerability=dataclasses.replace(resist_all_vuln, sunburst=vulnerable),
        shell_model=normal_door_model,
        shell_color=Color(1, 1, 1, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "A massive burst of light energy may damage it.",
        ),
        map_icon=DoorMapIcon.Light,
        shield_model="sunburst",
    ),
    "SonicBoom": dock_type.BlastShieldDoorType(
        name="SonicBoom",
        vulnerability=dataclasses.replace(resist_all_vuln, imploder=vulnerable),
        shell_model=annihilator_door_model,
        shell_color=Color(1, 1, 1, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "A massive burst of light and dark energy may damage it.",
        ),
        map_icon=DoorMapIcon.Annihilator,
        shield_model="sonic_boom",
    ),
    "AgonEnergy": dock_type.PlanetaryEnergyDoorType(
        name="AgonEnergy",
        vulnerability=resist_all_vuln,
        shell_color=Color(0.64, 0.34, 0, 1),
        scan_text=(
            "There is a Luminoth barrier on the door blocking access. ",
            "Analysis indicates that the barrier is linked to the energy of Agon. "
            "Return the energy to the Agon Temple.",
        ),
        map_icon=DoorMapIcon.AgonEnergy,
        planetary_energy_item_id=-1,  # TODO
    ),
    "TorvusEnergy": dock_type.PlanetaryEnergyDoorType(
        name="TorvusEnergy",
        vulnerability=resist_all_vuln,
        shell_color=Color(0.31, 0.59, 38, 1),
        scan_text=(
            "There is a Luminoth barrier on the door blocking access. ",
            "Analysis indicates that the barrier is linked to the energy of Torvus. "
            "Return the energy to the Torvus Temple.",
        ),
        map_icon=DoorMapIcon.TorvusEnergy,
        planetary_energy_item_id=-1,  # TODO
    ),
    "SanctuaryEnergy": dock_type.PlanetaryEnergyDoorType(
        name="SanctuaryEnergy",
        vulnerability=resist_all_vuln,
        shell_color=Color(0.64, 0.34, 0, 1),
        scan_text=(
            "There is a Luminoth barrier on the door blocking access. ",
            "Analysis indicates that the barrier is linked to the energy of Sanctuary. "
            "Return the energy to the Sanctuary Temple.",
        ),
        map_icon=DoorMapIcon.AgonEnergy,
        planetary_energy_item_id=-1,  # TODO
    ),
    "EchoVisor": dock_type.EchoVisorDoorType(
        name="EchoVisor",
        vulnerability=normal_vuln,
        shell_model=dark_door_model,
        shell_color=Color(1 / 16, 1 / 16, 1 / 16, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Sonic detection gear needed to interface with this system. "
            "Neutralizing the control emitter may disable it.",
        ),
        map_icon=DoorMapIcon.EchoVisor,
        shield_model="echo_visor",
        visor_flags=VisorFlags.Echo,
        player_controller_proxy=8,
    ),
    "DarkVisor": dock_type.DarkVisorDoorType(
        name="DarkVisor",
        vulnerability=normal_vuln,
        shell_model=dark_door_model,
        shell_color=Color(1, 0, 0, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Scans indicate presence of a control system. "
            "Interface method unknown. Control units not present in the visible spectrum or current timespace. ",
        ),
        map_icon=DoorMapIcon.DarkVisor,
        shield_model="dark_visor",
        visor_flags=VisorFlags.Dark,
        player_controller_proxy=7,
    ),
    "Cannon": dock_type.BlastShieldDoorType(
        name="Cannon",
        vulnerability=dataclasses.replace(resist_all_vuln, cannon_ball=vulnerable_no_splash),
        shell_model=normal_door_model,
        shell_color=Color(1, 1 / 3, 0, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. The Cannon Ball may damage it.",
        ),
        map_icon=DoorMapIcon.Boost,
        shield_model="cannon_ball",
    ),
    "Charge": dock_type.BlastShieldDoorType(
        name="Charge",
        vulnerability=dataclasses.replace(
            resist_all_vuln,
            power_charge=vulnerable,
            entangler=vulnerable,
            light_blast=vulnerable,
            sonic_boom=vulnerable,
        ),
        shell_model=normal_door_model,
        shell_color=Color(0, 1, 1, 1),
        scan_text=(
            "There is a Blast Shield on the door blocing access. ",
            "Analysis indicates that the Blast Shield is invulnerable to normal beam fire. "
            "The Charge Beam may damage it.",
        ),
        map_icon=DoorMapIcon.Normal,
        shield_model="charge_beam",
    ),
    "Power": dock_type.NormalDoorType(
        name="Power",
        vulnerability=dataclasses.replace(resist_all_vuln, power=vulnerable),
        shell_color=Color(1, 0.8, 0, 1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to light and dark energy. "
            "The Power Beam may damage it.",
        ),
        map_icon=DoorMapIcon.PowerBomb,  # TODO: give it its own?
    ),
}



import argparse
import importlib
import logging
import logging.config
import sys

_game_to_patcher = {
    "echoes": "open_prime_rando.echoes.cli",
}


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("game", choices=sorted(_game_to_patcher.keys()))
    return parser


def setup_logging():
    handlers = {
        "default": {
            "level": "DEBUG",
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",  # Default is stderr
        },
    }
    logging.config.dictConfig(
        {
            "version": 1,
            "formatters": {
                "default": {
                    "format": "[%(asctime)s] [%(levelname)s] [%(name)s] %(funcName)s: %(message)s",
                }
            },
            "handlers": handlers,
            "disable_existing_loggers": False,
            "loggers": {
                "default": {
                    "level": "DEBUG",
                },
            },
            "root": {
                "level": "DEBUG",
                "handlers": list(handlers.keys()),
            },
        }
    )


def main():
    setup_logging()
    parser = create_parser()
    args = parser.parse_args(sys.argv[1:2])

    patcher_module = importlib.import_module(_game_to_patcher[args.game])
    patcher_module.main(sys.argv[2:])



from retro_data_structures.formats.mrea import Area
from retro_data_structures.formats.strg import Strg
from retro_data_structures.properties.echoes.objects.WorldTeleporter import WorldTeleporter

from open_prime_rando.patcher_editor import PatcherEditor


def patch_elevator(
    editor: PatcherEditor,
    area: Area,
    elevator_id: int,
    target_mlvl: int,
    target_mrea: int,
    target_strg: int,
    target_name: str,
):
    elevator = area.get_instance(elevator_id)
    with elevator.edit_properties(WorldTeleporter) as props:
        props.world = target_mlvl
        props.area = target_mrea

    if target_strg is not None:
        strg = editor.get_file(target_strg, Strg)
        strg.set_single_string(
            1,
            f"Access to &push;&main-color=#FF3333;{target_name} &pop;granted. "
            f"Step into the hologram to activate elevator.",
        )



from __future__ import annotations

import logging
import uuid
from typing import TYPE_CHECKING

import open_prime_rando_practice_mod
from ppc_asm.assembler import ppc
from retro_data_structures.asset_manager import IsoFileProvider, PathFileWriter
from retro_data_structures.exceptions import UnknownAssetId
from retro_data_structures.game_check import Game
from retro_data_structures.properties.echoes.objects import WorldTeleporter

from open_prime_rando import practice_mod
from open_prime_rando.dol_patching import ppc_helper
from open_prime_rando.dol_patching.echoes import dol_patcher
from open_prime_rando.dol_patching.echoes.beam_configuration import BeamAmmoConfiguration
from open_prime_rando.dol_patching.echoes.user_preferences import OprEchoesUserPreferences
from open_prime_rando.echoes import custom_assets, frontend_asset_ids, inverted
from open_prime_rando.echoes.elevators import auto_enabled_elevator_patches
from open_prime_rando.patcher_editor import PatcherEditor

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from open_prime_rando.dol_patching.echoes.dol_patches import EchoesDolVersion
    from open_prime_rando.echoes.rando_configuration import AreaReference, RandoConfiguration

LOG = logging.getLogger("echoes_patcher")


def _default_dol_patches() -> dol_patcher.EchoesDolPatchesData:
    return dol_patcher.EchoesDolPatchesData(
        world_uuid=uuid.UUID("00000000-0000-1111-0000-000000000000"),
        energy_per_tank=100,
        beam_configurations=[
            BeamAmmoConfiguration.from_json(it)
            for it in [
                {
                    "item_index": 0,
                    "ammo_a": -1,
                    "ammo_b": -1,
                    "uncharged_cost": 0,
                    "charged_cost": 0,
                    "combo_missile_cost": 5,
                    "combo_ammo_cost": 0,
                },
                {
                    "item_index": 1,
                    "ammo_a": 45,
                    "ammo_b": -1,
                    "uncharged_cost": 1,
                    "charged_cost": 5,
                    "combo_missile_cost": 5,
                    "combo_ammo_cost": 30,
                },
                {
                    "item_index": 2,
                    "ammo_a": 46,
                    "ammo_b": -1,
                    "uncharged_cost": 1,
                    "charged_cost": 5,
                    "combo_missile_cost": 5,
                    "combo_ammo_cost": 30,
                },
                {
                    "item_index": 3,
                    "ammo_a": 46,
                    "ammo_b": 45,
                    "uncharged_cost": 1,
                    "charged_cost": 5,
                    "combo_missile_cost": 5,
                    "combo_ammo_cost": 30,
                },
            ]
        ],
        safe_zone_heal_per_second=1.0,
        user_preferences=OprEchoesUserPreferences(),
        default_items={"visor": "Combat Visor", "beam": "Power Beam"},
        unvisited_room_names=True,
        teleporter_sounds=True,
        dangerous_energy_tank=False,
    )


_ALL_FEATURES = False


def edit_starting_area_dol(editor: PatcherEditor, version: EchoesDolVersion, starting_area: AreaReference) -> None:
    function_address = version.starting_area_serialize_clean_slot_address  # CGameState::SerializeNewForCleanSlot

    # SetMLvlId argument
    ppc_helper.load_32bit_int(
        editor.dol,
        ppc.r4,
        starting_area.mlvl_id,
        function_address + 12 * 4,
        function_address + 14 * 4,
    )

    # StateForWorld argument
    ppc_helper.load_32bit_int(
        editor.dol,
        ppc.r4,
        starting_area.mlvl_id,
        function_address + 16 * 4,
        function_address + 18 * 4,
    )

    # SetDesiredAreaAssetId
    ppc_helper.load_32bit_int(
        editor.dol,
        ppc.r4,
        starting_area.mrea_id,
        function_address + 20 * 4,
        function_address + 21 * 4,
    )


def edit_starting_area_teleporter(editor: PatcherEditor, starting_area: AreaReference) -> None:
    try:
        area = editor.get_area(frontend_asset_ids.FRONTEND_PAL_MLVL, frontend_asset_ids.FRONTEND_PAL_MREA)
    except UnknownAssetId:
        area = editor.get_area(frontend_asset_ids.FRONTEND_NTSC_MLVL, frontend_asset_ids.FRONTEND_NTSC_MREA)

    elevator = area.get_instance("StartNewSinglePlayerGame")
    with elevator.edit_properties(WorldTeleporter) as teleporter:
        teleporter.world = starting_area.mlvl_id
        teleporter.area = starting_area.mrea_id


def edit_starting_area(editor: PatcherEditor, version: EchoesDolVersion, starting_area: AreaReference) -> None:
    edit_starting_area_dol(editor, version, starting_area)
    edit_starting_area_teleporter(editor, starting_area)


def patch_iso(
    input_iso: Path,
    output_iso: Path,
    configuration: RandoConfiguration,
    status_update: Callable[[str, float], None] = lambda s, _: LOG.info(s),
) -> None:
    """

    :param input_iso:
    :param output_iso:
    :param configuration:
    :param status_update:
    :return:
    """
    file_provider = IsoFileProvider(input_iso)

    output = PathFileWriter(output_iso)  # TODO: IsoFileWriter

    editor = PatcherEditor(file_provider, Game.ECHOES)

    custom_assets.create_custom_assets(editor)
    version = dol_patcher.apply_patches(editor.dol, _default_dol_patches())

    if configuration.starting_area is not None:
        edit_starting_area(editor, version, configuration.starting_area)

    if _ALL_FEATURES:
        auto_enabled_elevator_patches.apply_auto_enabled_elevators_patch(editor)
        inverted.apply_inverted(editor)

    if configuration.practice_mod != open_prime_rando_practice_mod.PracticeModMode.disabled:
        practice_mod.patch_dol(
            editor.dol,
            open_prime_rando_practice_mod.get_elf_for(version.practice_mod_version, configuration.practice_mod),
        )

    # Save our changes
    editor.build_modified_files()
    editor.save_modifications(output)
    status_update("Finished", 1.0)



import random
from dataclasses import dataclass

from retro_data_structures.base_resource import RawResource
from retro_data_structures.enums.echoes import Message, State
from retro_data_structures.formats.cmdl import Cmdl
from retro_data_structures.properties.echoes.archetypes.Transform import Transform
from retro_data_structures.properties.echoes.core.Vector import Vector
from retro_data_structures.properties.echoes.objects.Actor import Actor
from retro_data_structures.properties.echoes.objects.DamageableTrigger import DamageableTrigger
from retro_data_structures.properties.echoes.objects.Waypoint import Waypoint

from open_prime_rando.echoes.asset_ids.sanctuary_fortress import MAIN_GYRO_CHAMBER_MREA
from open_prime_rando.echoes.asset_ids.world import SANCTUARY_FORTRESS_MLVL
from open_prime_rando.echoes.custom_assets import custom_asset_path
from open_prime_rando.patcher_editor import PatcherEditor

RUBIKS_CUBES = {
    "Puzzle 1": [
        0x240013,
        0x2401A4,
        0x240145,
        0x24010A,
        0x24000E,
        0x240192,
        0x240139,
        0x24013A,
        0x240148,
    ],
    "Puzzle 2": [
        0x2402BA,
        0x2402AD,
        0x2402A6,
        0x2401C6,
        0x2401BE,
        0x2401C0,
        0x24028B,
        0x2401C3,
        0x240290,
    ],
}


@dataclass(frozen=True)
class RubiksColor:
    name: str
    state: State
    cmdl: int
    old_txtr: int

    @property
    def txtr_name(self) -> str:
        return f"rubiks_{self.name.lower()}.TXTR"

    @property
    def txtr(self) -> RawResource:
        return RawResource(
            type="TXTR",
            data=custom_asset_path().joinpath("rubiks", self.txtr_name).read_bytes(),
        )


COLORS = {
    "RED": RubiksColor(
        name="red",
        state=State.InternalState00,
        cmdl=0xF21AB8BF,
        old_txtr=0x29A5FF4B,
    ),
    "GREEN": RubiksColor(
        name="green",
        state=State.InternalState01,
        cmdl=0x5F2ADFC3,
        old_txtr=0xEE2889A3,
    ),
    "BLUE": RubiksColor(
        name="blue",
        state=State.InternalState02,
        cmdl=0x8F25F715,
        old_txtr=0xFED23B81,
    ),
}


def randomize_rubiks_puzzles(editor: PatcherEditor, rng: random.Random):
    area = editor.get_area(SANCTUARY_FORTRESS_MLVL, MAIN_GYRO_CHAMBER_MREA)

    # Add custom textures so colorblind players can distinguish the cubes
    for color in COLORS.values():
        txtr_id = editor.add_new_asset(color.txtr_name, color.txtr)

        cmdl = editor.get_file(color.cmdl, Cmdl)
        file_ids = cmdl.raw.material_sets[0].texture_file_ids
        old_txtr = file_ids.index(color.old_txtr)
        file_ids[old_txtr] = txtr_id
        editor._clear_cached_dependencies_for_asset(color.cmdl)  # FIXME: this should be done by RDS

    for puzzle_name, cubes in RUBIKS_CUBES.items():
        solution = [
            COLORS["RED"],
            COLORS["RED"],
            COLORS["RED"],
            COLORS["GREEN"],
            COLORS["GREEN"],
            COLORS["GREEN"],
            COLORS["BLUE"],
            COLORS["BLUE"],
            COLORS["BLUE"],
        ]
        rng.shuffle(solution)

        puzzle = area.get_instance(puzzle_name)
        for color, cube_id in zip(solution, cubes):
            cube = area.get_instance(cube_id)

            puzzle.remove_connections_from(cube)
            puzzle.add_connection(color.state, Message.Attach, cube)

            with cube.edit_properties(Actor) as props:
                props.model = color.cmdl

    patch_upstairs_puzzle_transform(editor)

    area.update_all_dependencies()


def patch_upstairs_puzzle_transform(editor: PatcherEditor):
    """
    The vanilla transform for these angles them with the glass, making the lock-on points
    awfully inconsistent. This patches them to be straight up and down, and adjusts the
    in/out lasers to match properly.
    """
    area = editor.get_area(SANCTUARY_FORTRESS_MLVL, MAIN_GYRO_CHAMBER_MREA)

    cubes = tuple(RUBIKS_CUBES["Puzzle 2"])
    # inner lasers
    lasers = (0x2402A9, 0x24028D, 0x240293, 0x2402A7, 0x2402B4, 0x2402B0)
    # buttons
    buttons = (0x2401C9, 0x2401C7, 0x2402B9, 0x2402AE)
    # damageable triggers
    damageable = (0x240294, 0x2402B2, 0x2402A8, 0x2401BC)

    Y_POS = 249.3522

    for actor_id in cubes + lasers:
        with area.get_instance(actor_id).edit_properties(Actor) as actor:
            actor.editor_properties.transform.rotation.x = 0.0
            actor.editor_properties.transform.position.y = Y_POS
            actor.is_solid = True

    for button_id in buttons:
        with area.get_instance(button_id).edit_properties(Actor) as actor:
            actor.editor_properties.transform.rotation.x = 0.0
            actor.editor_properties.transform.position.y = Y_POS

    for actor_id in damageable:
        with area.get_instance(actor_id).edit_properties(DamageableTrigger) as actor:
            actor.editor_properties.transform.rotation.x = 0.0
            actor.editor_properties.transform.position.y = Y_POS

    # axis ref. this needs to be updated so the puzzle rotates properly
    with area.get_instance(0x2402A5).edit_properties(Waypoint) as axis_ref:
        axis_ref.editor_properties.transform.rotation.x = 0.0

    big_lasers = {
        # red in
        0x2401C1: Transform(
            position=Vector(60.9426, 250.63, 12.0627), rotation=Vector(0.0, 0.1, 164.48), scale=Vector(0.9, 1.0, 1.0)
        ),
        # red out
        0x24028A: Transform(
            position=Vector(76.1287, 250.63, 12.0627), rotation=Vector(0.0, -0.7, -164.08), scale=Vector(0.9, 1.0, 1.0)
        ),
        # green in
        0x2401C2: Transform(
            position=Vector(60.342598, 250.13, 12.3027), rotation=Vector(0.0, 0.5, -187.5), scale=Vector(1.0, 1.0, 1.0)
        ),
        # green out
        0x2402BB: Transform(
            position=Vector(76.62867, 250.11, 12.2827), rotation=Vector(0.0, -0.5, 188.0), scale=Vector(0.9, 1.0, 1.0)
        ),
        # blue in
        0x24028F: Transform(
            position=Vector(60.142597, 249.51, 12.5127), rotation=Vector(0.0, 0.5, -180.0), scale=Vector(1.0, 1.0, 1.0)
        ),
        # blue out
        0x2402B6: Transform(
            position=Vector(77.1287, 249.51, 12.5127), rotation=Vector(0.0, -0.5, -180.0), scale=Vector(1.0, 1.0, 1.0)
        ),
    }

    for laser_id, transform in big_lasers.items():
        with area.get_instance(laser_id).edit_properties(Actor) as laser:
            laser.editor_properties.transform = transform
            laser.is_solid = True



from PyInstaller.utils.hooks import collect_data_files

import open_prime_rando.echoes.asset_ids.world

# https://pyinstaller.readthedocs.io/en/stable/hooks.html#provide-hooks-with-package

datas = collect_data_files("open_prime_rando", excludes=["__pyinstaller"])
hiddenimports = [
    f"open_prime_rando.echoes.asset_ids{module_name}"
    for module_name in open_prime_rando.echoes.asset_ids.world._DEDICATED_FILES.values()
]



import pytest

from open_prime_rando.echoes import patcher
from open_prime_rando.echoes.rando_configuration import PracticeModMode, RandoConfiguration


@pytest.fixture
def configuration(test_files_dir):
    data = test_files_dir.joinpath("echoes", "new_patcher.json").read_text()
    return RandoConfiguration.model_validate_json(data)


def test_ntsc_export(prime2_ntsc_iso_path, tmp_path, configuration) -> None:
    patcher.patch_iso(prime2_ntsc_iso_path, tmp_path, configuration)


def test_pal_export(prime2_pal_iso_path, tmp_path, configuration) -> None:
    configuration.practice_mod = PracticeModMode.disabled
    patcher.patch_iso(prime2_pal_iso_path, tmp_path, configuration)



import json
import logging
import typing
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING

from retro_data_structures.asset_manager import FileProvider, PathFileWriter
from retro_data_structures.formats.strg import Strg
from retro_data_structures.game_check import Game

from open_prime_rando.echoes import (
    asset_ids,
    dock_lock_rando,
    legacy_dynamic_schema,
)
from open_prime_rando.echoes.elevators.elevator_rando import patch_elevator
from open_prime_rando.echoes.general_changes import apply_corrupted_memory_card_change
from open_prime_rando.echoes.small_randomizations import apply_small_randomizations
from open_prime_rando.echoes.specific_area_patches import required_fixes
from open_prime_rando.echoes.suit_cosmetics import apply_custom_suits
from open_prime_rando.patcher_editor import PatcherEditor
from open_prime_rando.validator_with_default import DefaultValidatingDraft7Validator

if TYPE_CHECKING:
    from retro_data_structures.formats.mrea import Area

LOG = logging.getLogger("echoes_patcher")


def _read_legacy_schema():
    with Path(__file__).parent.joinpath("legacy_schema.json").open() as f:
        return json.load(f)


def apply_area_modifications(
    editor: PatcherEditor, configuration: dict[str, dict], status_update: Callable[[str, float], None]
):
    num_areas = sum(len(world_config["areas"]) for world_config in configuration.values())
    areas_processed = 0.0

    for world_name, world_config in configuration.items():
        world_meta = asset_ids.world.load_dedicated_file(world_name)
        mlvl = editor.get_mlvl(asset_ids.world.NAME_TO_ID_MLVL[world_name])

        mrea_to_name: dict[int, str] = {mrea: name for name, mrea in world_meta.NAME_TO_ID_MREA.items()}

        areas_by_name: dict[str, Area] = {mrea_to_name[area.mrea_asset_id]: area for area in mlvl.areas}

        for i, (area_name, area) in enumerate(areas_by_name.items()):
            if area_name not in world_config["areas"]:
                continue

            status_update(f"Processing {world_name} - {area_name}...", areas_processed / num_areas)
            areas_processed += 1

            area_config = world_config["areas"][area_name]
            low_memory = area_config["low_memory_mode"]

            for dock_name, dock_config in area_config["docks"].items():
                if "new_door_type" in dock_config:
                    dock_lock_rando.apply_door_rando(
                        editor,
                        world_name,
                        area_name,
                        dock_name,
                        dock_config["new_door_type"],
                        dock_config.get("old_door_type"),
                        low_memory,
                    )

            for layer_name, layer_state in area_config["layers"].items():
                LOG.debug("Setting layer %s of %s - %s to %s", layer_name, world_name, area_name, str(layer_state))
                area.get_layer(layer_name).active = layer_state

            for elevator in area_config["elevators"]:
                patch_elevator(
                    editor,
                    area,
                    elevator["instance_id"],
                    elevator["target_assets"]["world_asset_id"],
                    elevator["target_assets"]["area_asset_id"],
                    elevator["target_strg"],
                    elevator["target_name"],
                )

            if area_config["new_name"] is not None:
                old_strg = area._raw.area_name_id
                new_strg_id = editor.duplicate_asset(old_strg, f"custom_name_for_{area.internal_name}.STRG")
                strg = editor.get_file(new_strg_id, Strg)
                strg.set_single_string(0, area_config["new_name"])
                area._raw.area_name_id = new_strg_id

            area.update_all_dependencies(only_modified=True)


def apply_tweak_edits(editor: PatcherEditor, tweak_edits: dict[str, dict[str, typing.Any]]) -> None:
    """
    Edits the tweaks based on the generic schema api
    :param editor:
    :param tweak_edits:
    :return:
    """
    for instance in editor.tweaks.instances:
        properties = instance.get_properties().to_json()
        if properties["instance_name"] in tweak_edits:
            logging.debug("Editing %s", properties["instance_name"])

            for name, value in tweak_edits[properties["instance_name"]].items():
                parent = properties
                spit_name = name.split(".")

                for part in spit_name[:-1]:
                    parent = parent[part]

                parent[spit_name[-1]] = value

            instance.set_properties(instance.type.from_json(properties))


def patch_paks(
    file_provider: FileProvider,
    output_path: Path,
    configuration: dict,
    status_update: Callable[[str, float], None] = lambda s, _: LOG.info(s),
):
    """Applies the legacy patches, intended to be used alongside Claris' patcher."""
    status_update(f"Will patch files at {file_provider}", 0)
    output_path.joinpath("files").mkdir(parents=True, exist_ok=True)
    output_path.joinpath("files", "opr_patcher_data.json").write_text(json.dumps(configuration))

    editor = PatcherEditor(file_provider, Game.ECHOES)

    status_update("Preparing schema", 0)
    schema = legacy_dynamic_schema.expand_schema(_read_legacy_schema(), editor)

    status_update("Validating schema", 0)
    DefaultValidatingDraft7Validator(schema).validate(configuration)

    status_update("Applying small patches", 0)
    dock_lock_rando.add_custom_models(editor)
    required_fixes.torvus_temple(editor)
    required_fixes.command_center_door(editor)
    apply_small_randomizations(editor, configuration["small_randomizations"])
    apply_corrupted_memory_card_change(editor)

    if "tweaks" in configuration:
        status_update("Modifying tweaks", 0)
        apply_tweak_edits(editor, configuration["tweaks"])

    status_update("Modifying areas", 0)
    apply_area_modifications(editor, configuration["worlds"], status_update)
    apply_custom_suits(editor, configuration["cosmetics"]["suits"])

    # Save our changes
    editor.build_modified_files()

    editor.save_modifications(PathFileWriter(output_path))
    status_update("Finished", 1.0)



from unittest.mock import MagicMock, call

from open_prime_rando.echoes import custom_assets


def test_create_split_ammo():
    manager = MagicMock()

    custom_assets._create_split_ammo(manager)

    # Assert
    manager.duplicate_asset.assert_has_calls(
        [
            call(custom_assets.BEAM_AMMO_EXPANSION_CMDL, "dark_ammo.CMDL"),
            call(custom_assets.BEAM_AMMO_EXPANSION_ANCS, "dark_ammo.ANCS"),
            call(custom_assets.BEAM_AMMO_EXPANSION_CMDL, "light_ammo.CMDL"),
            call(custom_assets.BEAM_AMMO_EXPANSION_ANCS, "light_ammo.ANCS"),
        ]
    )



import random
from enum import Enum

from retro_data_structures.enums.echoes import Message, State
from retro_data_structures.formats.strg import Strg

from open_prime_rando.echoes.asset_ids.sanctuary_fortress import MINIGYRO_CHAMBER_MREA
from open_prime_rando.echoes.asset_ids.world import SANCTUARY_FORTRESS_MLVL
from open_prime_rando.patcher_editor import PatcherEditor


class GyroColor(Enum):
    AMBER = 0
    COBALT = 1
    CRIMSON = 2
    EMERALD = 3

    @property
    def color(self) -> str:
        if self == GyroColor.AMBER:
            return "#A45600"
        if self == GyroColor.COBALT:
            return "#56789D"
        if self == GyroColor.CRIMSON:
            return "#FF6562"
        if self == GyroColor.EMERALD:
            return "#4E9761"

    @property
    def text(self) -> str:
        return f"&push;&main-color={self.color};{self.name}&pop;"


GYRO_STATES = [
    State.InternalState00,
    State.InternalState01,
    State.InternalState02,
    State.InternalState03,
]


def randomize_minigyro_chamber(editor: PatcherEditor, rng: random.Random):
    area = editor.get_area(SANCTUARY_FORTRESS_MLVL, MINIGYRO_CHAMBER_MREA)
    solution = [GyroColor.AMBER, GyroColor.COBALT, GyroColor.CRIMSON, GyroColor.EMERALD]
    rng.shuffle(solution)

    counter = area.get_instance("Stage gate activator")
    stage_gates = [area.get_instance(f"Stage gate {i + 1}") for i in range(4)]

    for i, gate in enumerate(stage_gates):
        counter.remove_connections_from(gate)
        for j, gyro in enumerate(solution):
            message = Message.Activate if i == gyro.value else Message.Deactivate
            counter.add_connection(GYRO_STATES[j], message, gate)

    # play jingle on the final gyro
    stop_gyros = [area.get_instance(f"[IN/OUT] Stop gyroscope {i + 1}") for i in range(4)]
    jingle = area.get_instance("StreamedAudio - Event Jingle")

    stop_gyros[3].remove_connections_from(jingle)
    stop_gyros[solution[3].value].add_connection(State.Zero, Message.Play, jingle)

    scan = editor.get_file(0xFBFF349D, Strg)
    solution_text = "\n".join(gyro.text for gyro in solution)
    scan.set_single_string(1, f"Safety lockdown code is as follows:\n\n\n{solution_text}")

    area.update_all_dependencies()



import dataclasses

from retro_data_structures.properties.echoes.archetypes.DamageVulnerability import DamageVulnerability
from retro_data_structures.properties.echoes.archetypes.WeaponVulnerability import Effect, WeaponVulnerability

reflect = WeaponVulnerability(damage_multiplier=0, effect=Effect.Reflect)
vulnerable = WeaponVulnerability(damage_multiplier=100, effect=Effect.Normal)
immune = WeaponVulnerability(damage_multiplier=0, effect=Effect.Normal, ignore_radius=True)
vulnerable_no_splash = WeaponVulnerability(damage_multiplier=100.0, effect=Effect.Normal, ignore_radius=True)


resist_all_vuln = DamageVulnerability(
    power=reflect,
    dark=reflect,
    light=reflect,
    annihilator=reflect,
    power_charge=reflect,
    entangler=reflect,
    light_blast=reflect,
    sonic_boom=reflect,
    super_missle=reflect,
    black_hole=reflect,
    sunburst=reflect,
    imploder=reflect,
    boost_ball=immune,
    cannon_ball=immune,
    screw_attack=immune,
    bomb=immune,
    power_bomb=immune,
    missile=reflect,
    phazon=reflect,
    ai=immune,
    poison_water=immune,
    dark_water=immune,
    lava=immune,
    area_damage_hot=immune,
    area_damage_cold=immune,
    area_damage_dark=immune,
    area_damage_light=immune,
    weapon_vulnerability=immune,
    normal_safe_zone=immune,
)

normal_vuln = dataclasses.replace(
    resist_all_vuln,
    power=vulnerable,
    dark=vulnerable,
    light=vulnerable,
    annihilator=vulnerable,
    power_charge=vulnerable,
    entangler=vulnerable,
    light_blast=vulnerable,
    sonic_boom=vulnerable,
    super_missle=vulnerable,
    black_hole=vulnerable,
    sunburst=vulnerable,
    imploder=vulnerable,
    missile=vulnerable,
    bomb=vulnerable,
    power_bomb=vulnerable,
)



from open_prime_rando.dol_patching.prime1.dol_patches import Prime1DolVersion

ALL_VERSIONS = [
    Prime1DolVersion(
        version="0-00",
        description="GC NTSC 0-00",
        build_string_address=0x803CC588,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.088 10/29/2002 2:21:25",
        sda2_base=0x805B1D20,
        sda13_base=0x805AEBC0,
        cplayer_vtable=0x803D96E8,
        message_receiver_string_ref=0x803EFB90,
    ),
    Prime1DolVersion(
        version="0-01",
        description="GC NTSC 0-01",
        build_string_address=0x803CC768,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.093 11/5/2002 19:50:01",
        sda2_base=0x805B1F00,
        sda13_base=0x805AEDA0,
        cplayer_vtable=0x803D98C8,
        message_receiver_string_ref=0x803EFD70,
    ),
    Prime1DolVersion(
        version="0-02",
        description="GC NTSC 0-02",
        build_string_address=0x803CD648,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.111 3/10/2003 17:56:21",
        sda2_base=0x805B2DE0,
        sda13_base=0x805AFC40,
        cplayer_vtable=0x803DA7A8,
        message_receiver_string_ref=0x803F0BA8,
    ),
    Prime1DolVersion(
        version="pal",
        description="GC PAL",
        build_string_address=0x803B6924,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.110 2/4/2003 22:16:07",
        sda2_base=0x80473E60,
        sda13_base=0x80470C60,
        cplayer_vtable=0x803C4B88,
        message_receiver_string_ref=0x803D7A28,
    ),
    Prime1DolVersion(
        version="kor",
        description="GC NTSC-K",
        build_string_address=0x803CC688,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.097 12/19/2002 16:03:43",
        sda2_base=0x805B1A00,
        sda13_base=0x805AE8A0,
        cplayer_vtable=0x803D97E8,
        message_receiver_string_ref=0x803EFC90,
    ),
    Prime1DolVersion(
        version="jpn",
        description="GC NTSC-J",
        build_string_address=0x803B86CC,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.111 2/11/2003 10:03:22",
        sda2_base=0x8059AE00,
        sda13_base=0x80597C20,
        cplayer_vtable=0x803C5B28,
        message_receiver_string_ref=0x803D89C8,
    ),
]



import open_prime_rando_practice_mod
from retro_data_structures.game_check import Game

from open_prime_rando.dol_patching.all_prime_dol_patches import (
    DangerousEnergyTankAddresses,
    HealthCapacityAddresses,
    PowerupFunctionsAddresses,
    StringDisplayPatchAddresses,
)
from open_prime_rando.dol_patching.echoes.dol_patches import (
    BeamCostAddresses,
    EchoesDolVersion,
    IsDoorAddr,
    MapDoorTypeAddresses,
    SafeZoneAddresses,
    StartingBeamVisorAddresses,
)

ALL_VERSIONS = [
    EchoesDolVersion(
        game=Game.ECHOES,
        description="Gamecube NTSC",
        build_string_address=0x803AC3B0,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.028 10/18/2004 10:44:32",
        sda2_base=0x804223C0,
        sda13_base=0x8041FD80,
        game_state_pointer=0x80418EB8,
        cplayer_vtable=0x803B15D0,
        cstate_manager_global=0x803DB6E0,
        string_display=StringDisplayPatchAddresses(
            update_hint_state=0x80038020,
            message_receiver_string_ref=0x803BD118,
            wstring_constructor=0x802FF3DC,
            display_hud_memo=0x8006B3C8,
            max_message_size=200,
        ),
        powerup_functions=PowerupFunctionsAddresses(
            add_power_up=0x800858F0,
            incr_pickup=0x80085760,
            decr_pickup=0x800856C4,
        ),
        practice_mod_version=open_prime_rando_practice_mod.GameVersion.GC_NTSC,
        health_capacity=HealthCapacityAddresses(
            base_health_capacity=0x8041ABE4,
            energy_tank_capacity=0x8041ABE0,
        ),
        dangerous_energy_tank=DangerousEnergyTankAddresses(
            small_number_float=0x8041A4A8,
            incr_pickup=0x80085760,
        ),
        beam_cost_addresses=BeamCostAddresses(
            uncharged_cost=0x803AA8C8,
            charged_cost=0x803AA8D8,
            charge_combo_ammo_cost=0x803AA8E8,
            charge_combo_missile_cost=0x803A74AC,
            get_beam_ammo_type_and_costs=0x801CCCB0,
            is_out_of_ammo_to_shoot=0x801C92E8,
            gun_get_player=0x801DD758,
            get_item_amount=0x80085514,
        ),
        game_options_constructor_address=0x80161B48,
        safe_zone=SafeZoneAddresses(
            heal_per_frame_constant=0x8041A4FC,
            increment_health_fmr=0x8000C710,
        ),
        starting_beam_visor=StartingBeamVisorAddresses(
            player_state_constructor_clean=0x80086008,
            player_state_constructor_decode=0x80085C20,
            health_info_constructor=0x801420C8,
            enter_morph_ball_state=0x80184118,
            start_transition_to_visor=0x80085314,
            reset_visor=0x80085334,
        ),
        anything_set_address=0x8010F084,
        rs_debugger_printf_loop_address=0x8028C604,
        unvisited_room_names_address=0x8008B714,
        cworldtransmanager_sfxstart=0x80158E50,
        powerup_should_persist=0x803A743C,
        map_door_types=MapDoorTypeAddresses(
            get_correct_transform=IsDoorAddr(0x800BB4D8, 0x800BB4E0, 3),
            map_obj_draw=IsDoorAddr(0x800BB960, 0x800BB96C, 7),
            is_visible_to_automapper=IsDoorAddr(0x800BB600, 0x800BB608, 3),
            map_world_draw_areas=IsDoorAddr(0x8009458C, 0x80094594, 0),
            map_area_commit_resources1=IsDoorAddr(0x8007F3B4, 0x8007F3BC, 3),
            map_area_commit_resources2=IsDoorAddr(0x8007FAB0, 0x8007FAB8, 3),
            get_door_color=0x802175B4,
            map_icon_jumptable=0x803B3638,
        ),
        double_damage_vfx=0x80017F28,
        starting_area_serialize_clean_slot_address=0x80143548,
    ),
    EchoesDolVersion(
        game=Game.ECHOES,
        description="Gamecube PAL",
        build_string_address=0x803AD710,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.035 10/27/2004 19:48:17",
        sda2_base=0x804236C0,
        sda13_base=0x80421060,
        game_state_pointer=0x8041A19C,
        cplayer_vtable=0x803B2950,
        cstate_manager_global=0x803DC900,
        string_display=StringDisplayPatchAddresses(
            update_hint_state=0x80038194,
            message_receiver_string_ref=0x803BE378,
            wstring_constructor=0x802FF734,
            display_hud_memo=0x8006B504,
            max_message_size=200,
        ),
        powerup_functions=PowerupFunctionsAddresses(
            add_power_up=0x80085A2C,
            incr_pickup=0x8008589C,
            decr_pickup=0x80085800,
        ),
        practice_mod_version=open_prime_rando_practice_mod.GameVersion.GC_PAL,
        health_capacity=HealthCapacityAddresses(
            base_health_capacity=0x8041BEDC,
            energy_tank_capacity=0x8041BED8,
        ),
        dangerous_energy_tank=DangerousEnergyTankAddresses(
            small_number_float=0x8041B7A0,
            incr_pickup=0x8008589C,
        ),
        beam_cost_addresses=BeamCostAddresses(
            uncharged_cost=0x803ABC28,
            charged_cost=0x803ABC38,
            charge_combo_ammo_cost=0x803ABC48,
            charge_combo_missile_cost=0x803A7C04,
            get_beam_ammo_type_and_costs=0x801CCFE4,
            is_out_of_ammo_to_shoot=0x801C961C,
            gun_get_player=0x801DDA8C,
            get_item_amount=0x80085650,
        ),
        game_options_constructor_address=0x80161D9C,
        safe_zone=SafeZoneAddresses(
            heal_per_frame_constant=0x8041B7F4,
            increment_health_fmr=0x8000C754,
        ),
        starting_beam_visor=StartingBeamVisorAddresses(
            player_state_constructor_clean=0x80086144,
            player_state_constructor_decode=0x80085D5C,
            health_info_constructor=0x80142304,
            enter_morph_ball_state=0x801843F8,
            start_transition_to_visor=0x80085450,
            reset_visor=0x80085470,
        ),
        anything_set_address=0x8010F238,
        rs_debugger_printf_loop_address=0x8028CA0C,
        unvisited_room_names_address=0x8008B850,
        cworldtransmanager_sfxstart=0x801590A4,
        powerup_should_persist=0x803A7B94,
        map_door_types=MapDoorTypeAddresses(
            get_correct_transform=IsDoorAddr(0x800BB56C, 0x800BB574, 3),
            map_obj_draw=IsDoorAddr(0x800BB9F0, 0x800BB9FC, 3),
            is_visible_to_automapper=IsDoorAddr(0x800BB694, 0x800BB69C, 3),
            map_world_draw_areas=IsDoorAddr(0x80094620, 0x80094628, 0),
            map_area_commit_resources1=IsDoorAddr(0x8007F4F0, 0x8007F4F8, 3),
            map_area_commit_resources2=IsDoorAddr(0x8007FBEC, 0x8007FBF4, 3),
            get_door_color=0x802178D4,
            map_icon_jumptable=0x803B4A80,
        ),
        double_damage_vfx=0x80017FC4,
        starting_area_serialize_clean_slot_address=0x80143784,
    ),
]
